#!/bin/bash
#
#Name:		saapsrun.sh
#Description:
#This is the SAAPS application launcher script.
#It includes a configuration part which is set at install time.
#To reconfigure SAAPS you can edit this script, please see "CONFIGURE HERE".
#
#The script findes the name it was called by and translates it to the java exec.
#
#Author:	Leif Kalla (leif@irf.se)

#Get the specific program that user wants to run
progname=`basename $0`
dirname=`dirname $0`

### CONFIGURE HERE ###

#If you want to have a static definition of the SAAPS_HOME
#modify the line below and uncomment it.
#Absolute path to SAAPS directory
#SAAPS_HOME=/space/someuser/saaps

#Models location
SAAPS_MODELS=${SAAPS_HOME}/Models/

#The location of the classes
SAAPS_CLASSES="${SAAPS_HOME}/lib/saaps.jar:${SAAPS_HOME}/lib/saapsExt.jar:${SAAPS_HOME}/lib/mysql-connector-java-5.1.10-bin.jar"

###SAAPS server stuff (RMI)
#Give the complete name for this computer (needed by the RMI server)
#SAAPS_MACHINE="butler.irf.se"
SAAPS_MACHINE="www.umea.irf.se"

#A string for the RMI server codebase
#Note: The classes doesn't need to be accessible to the applets
#through the RMI codebase as they already are distributed by the
#web server. The RMI codebase has to be set anyway.
SAAPS_RMI_CLASSES="file:${SAAPS_HOME}/lib/saaps.jar file:${SAAPS_HOME}/lib/saapsExt.jar"

#policy file 
SAAPS_POLICY="${SAAPS_HOME}/.mypolicy"

### END OF CONFIGURATION ###

### Verification of config parameters

#Checking the SAAPS_HOME setting
if [ ! -f ${SAAPS_HOME}/lib/saapsExt.jar ] && [ ${progname} != "respawnrmiserver" ] && [ $1 != "restart" ]
then
	echo "SAAPS_HOME=${SAAPS_HOME} is wrong!"
	echo "can't find lib/saapsExt.jar in ${SAAPS_HOME}"
	exit 1
fi 

### THE PART WHERE PROGRAM IS STARTED ###
#If any program is added to SAAPS and you want to start it with this script
#add an entry into the case clause

case "${progname}" in

    #The saaps RMI database server. NOTE! Do not forget to set codeBase path in the .mypolicy file.
    #Remember to open RMI tcp port 2001 on the servers firewall.
    #Tested browsers: Internet Explorer 8, Mozilla Firefox 3.5 and Google Chrome 6.
    saapsrmiserver)
	case "$1" in
	    start)
		echo "Starting RMI server"
		if [ -f ${SAAPS_HOME}/log/DatabaseRMI.pid ]
		then
			pid=`cat ${SAAPS_HOME}/log/DatabaseRMI.pid`
			ps -e|grep "$pid" >/dev/null 2>&1
			if [ $? -eq 0 ]
			then
				echo "Server already running"
				exit 0
			else
				rm -f ${SAAPS_HOME}/log/DatabaseRMI.pid
			fi
		fi
		mkdir -p /tmp/$USER
                echo ${SAAPS_HOME} >/tmp/$USER/SAAPS_HOME.path
		#start the java object registry
		# Now started internally in DatabaseRMIImpl()
#		rmiregistry 2001 &

		#Start our rmi server
		java -Djava.rmi.server.hostname=${SAAPS_MACHINE} \
		-Dsaaps.home=$SAAPS_HOME \
		-Dsaaps.models=$SAAPS_MODELS \
		-classpath $SAAPS_CLASSES \
		-Djava.security.policy=$SAAPS_POLICY \
		-Djava.rmi.server.logCalls=true \
		-Djava.rmi.server.codebase="${SAAPS_RMI_CLASSES}" \
		saaps.dbt.DatabaseRMIImpl \
		>> $SAAPS_HOME/log/DatabaseRMI.log 2>&1 &

		#Record the processid
		pid=$!
		echo "$pid" > $SAAPS_HOME/log/DatabaseRMI.pid
		;;

	    stop)
		echo "Stopping RMI server"
		if [ -f ${SAAPS_HOME}/log/DatabaseRMI.pid ]
		then
		    kill `cat ${SAAPS_HOME}/log/DatabaseRMI.pid`
		    rm -f "${SAAPS_HOME}/log/DatabaseRMI.pid"
		fi
		rm -f /tmp/$USER/SAAPS_HOME.path
		;;

		#script that checks if the SAAPS RMI server is running
	    check)
		if [ -f ${SAAPS_HOME}/log/DatabaseRMI.pid ]
		then
			pid=`cat ${SAAPS_HOME}/log/DatabaseRMI.pid`
			ps -e|grep "$pid" >/dev/null 2>&1
			if [ $? -eq 0 ]
			then
				echo "running"
				exit 0
			else
				echo "Server not running"
				exit 1
			fi
		else
			echo "Server not running"
			exit 1
		fi
		;;

		#script that force restart on SAAPS RMI server
	    restart)
		if [ -f /tmp/$LOGNAME/SAAPS_HOME.path ]; then
			export SAAPS_HOME=`cat /tmp/$LOGNAME/SAAPS_HOME.path`
			cd ${SAAPS_HOME}
			echo "Server Force Restart"
			./saapsrmiserver stop
			sleep 2
			./saapsrmiserver start
		fi
		;;

	    *)
		echo "Usage: saapsserver {start|stop|restart|check}" >&2
		exit 1
		;;
	esac
	;;

	#script that respawn the SAAPS RMI server
	respawnrmiserver)
		# crontab job example for check/respawn of the RMI server 6 times per hour
		# */10 * *   *   *  /home/someuser/odi-1.0/SAAPS/respawnrmiserver >>/home/someuser/odi-1.0/SAAPS/log/DatabaseRMI.log 2>&1
		# crontab job example for forced restart of RMI server every day at 4:57
		# 57   4 *   *   *  /home/someuser/odi-1.0/SAAPS/saapsrmiserver restart >>/home/someuser/odi-1.0/SAAPS/log/DatabaseRMI.log 2>&1
		#
		if [ -f /tmp/$LOGNAME/SAAPS_HOME.path ]; then
			export SAAPS_HOME=`cat /tmp/$LOGNAME/SAAPS_HOME.path`
			cd ${SAAPS_HOME}
			if [ -f ${SAAPS_HOME}/log/DatabaseRMI.pid ]
			then
				pid=`cat ${SAAPS_HOME}/log/DatabaseRMI.pid`
				ps -e|grep "$pid" >/dev/null 2>&1
				if [ $? -eq 0 ]
				then
					echo "running"
					exit 0
				else
					echo "Server not running. Restart"
					./saapsrmiserver start
					exit 1
				fi
			else
				echo "Server not running and no pid file. Restart"
				./saapsrmiserver start
				exit 1
			fi
		else
			echo Can not get SAAPS_HOME path. RMI server not respawned.
		fi
		;;

	#script to check that lib/saaps.jar is found in SAAPS_HOME
	checkpath)
		if [ -f ${SAAPS_HOME}/lib/saaps.jar ]
		then
			echo "SAAPS_HOME=${SAAPS_HOME} is correct! "
			exit 0
		else
			echo "SAAPS_HOME=${SAAPS_HOME} is wrong!"
			echo "can't find lib/saaps.jar in ${SAAPS_HOME}"
			exit 1
		fi
		;;

	#script to generate javadoc
	createapi)
		javadoc -private -windowtitle "SAAPS API DOC" -version -author \
		-doctitle "<H3><CENTER>Satellite Anomaly Analysis and Prediction System</CENTER></H3>" \
		-classpath  ${SAAPS_HOME}/lib/saapsExt.jar:${SAAPS_HOME}/ \
		-sourcepath ${SAAPS_HOME}/ \
		-d ${SAAPS_HOME}/web/api/.  \
		saaps saaps.ann saaps.beans saaps.dbt saaps.math saaps.network saaps.saam \
		saaps.sapm saaps.util
		;;

	*)
	    echo "No such program configured! If it's a new program add it into saapsrun.sh"
	    exit 1
	    ;;
esac

exit 0
