/*
 * Network.java
 *
 * Created on August 9, 2000, 3:10 PM
 */
 
package saaps.ann.mlff;

import saaps.ann.Data;
import java.io.Serializable;

/** 
 * This class implements the multi layer feed forward neural network.
 * The network is made up of one or several layers, which in turn contain
 * neurons. Each neuron is defined by its weights, bias, and transfer function.
 * <p>
 * The network can be stored to the file system and read from the file system.
 * <p>
 * To run the network one must first set the weights, biases, and transfer functions,
 * either explicitly or from reading a stored network. Then one specifies the input
 * as an <code>saaps.ann.Data</code> object. After it has been run the output
 * can be received as an <code>saaps.ann.Data</code> object.
 * <p>
 * Example:
 * <blockquote>
 *  This example assumes that we have an 2 element array of <code>Layer</code> objects
 *  called <code>layers</code>. The should also be a <code>Data</code> object called
 *  inData.
 *  <blockquote>
 *    Network net = new Network(layers);<br>
 *    net.setInput(inData);<br>
 *    Data outData = net.getOutput();<br>
 *  </blockquote>
 * </blockquote>
 *
 * @author  Peter Wintoft
 * @version 1.0 2000-08-11
 */
public class Network extends Object implements Serializable {

  //The network layers.
  private Layer[] layers;
  //The number of layers.
  private int numOfLayers;
  //The input data to the network.
  private transient Data networkInputData;
  //The output data from the network.
  private transient Data networkOutputData;
  //A flag indicating whether the network has been run for the input data.
  private transient boolean hasBeenRun = false;
  
  /** 
  * Creates a new empty Network
  */
  public Network() {
  }
  
  /**
  Create a new network by defining the number of inputs, the number of neurons in each layer,
  and the transfer functions of each layer.
  <p>
  Example:
  <p>
  Create a network with 10 inputs, 5 hidden, and 3 outputs. The hidden layer function is
  the tanh and the output layer function is linear.
  <blockquote>
  <tt>
  int numOfInputs = 10;<br>
  int[] numOfNeurons = {5,3};<br>
  String[] neurons = {"TanhNeuron","LinNeuron"};<br>
  Network net = new Network(numOfInputs,numOfNeurons,neurons);<br>
  </tt>
  </blockquote>
  
  @param  noi Number of inputs.
  @param  non Array with number of neurons in each layer.
  @param  f   Array with the name of the neurons.
  */
  public Network(int noi, int[] non, String[] f) {
    //Create the layers.
    numOfLayers = non.length;
    layers = new Layer[numOfLayers];
    for (int l=0;l<layers.length;l++) {
      //Create the neurons in each layer.
      Neuron[] neurons = new Neuron[non[l]];
      Neuron neuron = null;
      for (int n=0;n<neurons.length;n++) {
        try {
          Class classObject = Class.forName("saaps.ann.mlff." + f[l]);
          neuron = (Neuron) classObject.newInstance();
        }
        catch (ClassNotFoundException e) {
          System.err.println(e);
        }
        catch (InstantiationException e) {
          System.err.println(e);
        }
        catch (IllegalAccessException e) {
          System.err.println(e);
        }
        if (l==0)
          neuron.setNumOfInputs(noi);
        else
          neuron.setNumOfInputs(non[l-1]);
        neurons[n] = neuron;
      }
      layers[l] = new Layer(neurons);
    }
  }
        
  /**
  * Create a network from a stored file.
  * <p>
  * <bf>Not yet implemented!</bf>
  *
  * @param  name  The name of the file.
  */
  public Network(String name)
  {
  }
  
  /**
  Create a network from an array of layers.
  
  @param  l The layers
  */
  public Network(Layer[] l)
  {
    layers = l;
    numOfLayers = layers.length;
  }
  
  /**
  Get a layer from the network.
  
  @param  l The layer, with the first hidden layer having index 0.
  @return The layer.
  */
  public Layer getLayer(int l) {
    return layers[l];
  }
  
  /**
  Get the number of layers.
  
  @return The number of layers.
  */
  public int getNumOfLayers() {
    return layers.length;
  }
  
  /**
  * Present the input data to the network. The network is not run.
  *
  * @param  inp The input data.
  */
  public void setInput(Data inp)
  {
    networkInputData = inp;
    hasBeenRun = false;
  }
  
  /**
  * Run the network.
  */
  public void run()
  {
//    Data[] outputLayer = new Data[numOfLayers];
    Data input = networkInputData;
    Data output = input;
    for (int l=0;l<numOfLayers;l++)
    {
      layers[l].setInput(input);
      output = layers[l].getOutput();
      input = output;
    }
    networkOutputData = output;
    hasBeenRun = true;
  }
  
  /**
  * Get the output.
  * <p>
  * If the network has not been run after new
  * input data has been presented this method will also run the
  * network.
  *
  * @return The output from the network.
  */
  public Data getOutput()
  {
    if (hasBeenRun)
      return networkOutputData;
    else
    {
      run();
      return networkOutputData;
    }
  }
  
  public String toString() {
    StringBuffer buf = new StringBuffer(this.getClass().getName() + "\n");
    buf.append("Number of layers = " + numOfLayers + "\n");
    buf.append("Has been run = " + hasBeenRun + "\n");
    return buf.toString();
  }
}