/*
 * Neuron.java
 *
 * Created on August 9, 2000, 3:42 PM
 */
 
package saaps.ann.mlff;

import saaps.ann.Data;
import saaps.ann.Weights;

/** 
 * This class represents a neuron. The neuron contains the weights and a bias.
 * The class also contains an <code>abstract</code> method for the transfer function.
 * Thus, the class must be inherited by a subclass that implements the transfer function.
 * <p>
 * Example:
 *  <blockquote>
 *  double[] w = {-0.1,0.3,-0.5};<br>
 *  double b = 0.4;<br>
 *  Weights weights = new Weights(w,b);<br>
 *  TanhNeron myNeuron = new TanhNeuron(weights)<br>
 *  </blockquote>
 *
 * @author  Peter Wintoft
 * @version 0.1 2000-08-09
 * @version 1.0 2000-08-11
 * @see TanhNeuron
 * @see LinNeuron
 * @see Layer
 * @see Network
 */
public abstract class Neuron extends Object {

  //The weights and bias.
  private Weights weights;
  //The input data.
  private transient Data input;
  //The output data.
  private transient Data output;
  //A flag indicating whether the network has been run for the input data.
  private transient boolean hasBeenRun = false;
  
  /**
  * Creates new empty Neuron.
  */
  public Neuron() {
  }
 
  /**
  * Creates a neuron from a <code>Weights</code> object.
  *
  * @param  w The weights and bias
  */
  public Neuron(Weights w)
  {
    weights = w;
  }
  
  /**
  Get the neurons weights.
  
  @return The weights.
  */
  public Weights getWeights() {
    return weights;
  }
  
  /**
  Set the number of inputs to this neuron. The method also
  initializes the weights and bias to random numbers between
  -1 to +1.
  
  @param  n The number of inputs.
  */
  public void setNumOfInputs(int n) {
    weights = new Weights(n);
    weights.randomize(-1,1);
  }
  
  /**
  Set the input data to the neuron.
  
  @param  inp The input data
  */
  public void setInput(Data inp)
  {
    input = inp;
    hasBeenRun = false;
  }
  
  /**
  Get the output from the neuron.
  <p>
  If new input data has been presented to the neuron then the
  neuron is first run.
  
  @return The output data
  */
  public Data getOutput()
  {
    if (hasBeenRun)
      return output;
    else
    {
      run();
      return output;
    }
  }
  
  /**
  Run the neuron by performing the weighted sum over the inputs
  and then apply the transfer function.
  */
  private void run()
  {
    Data activation = input.weightedSum(weights);
    output = transFun(activation);
    hasBeenRun = true;
  }
  
  /**
  The transfer function.
  
  @param  in  The data dat goes into the transfer function
  @return The output from the tranfer function.
  */
  public abstract Data transFun(Data in);
  
  public String toString() {
    StringBuffer buf = new StringBuffer(this.getClass().getName() + "\n");
    buf.append("Number of inputs = " + weights.getLength() + "\n");
    buf.append("Has been run = " + hasBeenRun + "\n");
    return buf.toString();
  }

}