/*
 * BuildDst.java
 *
 * Created on March 5, 2001, 10:15 AM
 */
 
package saaps.dbt;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import java.util.Vector;

import saaps.util.IllegalTimeSeriesException;
import saaps.util.ObjectTimeSeries;

/** 
 * This class builds the AE database from NGDC.
 * The database consists of hourly values of AE, AL, AO, and AU.
 *
 * @author  Peter Wintoft
 * @version 1.0
 */
public class BuildAE extends Object {
    
    private String inputDir;
    private int yearFirst;
    private int yearLast;

  /** Creates new BuildAE */
  public BuildAE(String dir, int sy, int ey) {
      this.inputDir = dir;
      this.yearFirst = sy;
      this.yearLast = ey;
  }
  
  /**
  Go through the files with hourly values on the loacl hard disk.
  <p>
  Data exist for the years 1957 to 1988.
  */
  public void run() throws Exception {
    //First year and last year to read.
//    int yearFirst = 1957;
//    int yearLast = 1988;
    //Create the AEData object.
    AEDataOneYear aeData = new AEDataOneYear();
    //Go through the years.
    for (int year=yearFirst;year<=yearLast;year++) {
      System.out.println("Reading year " + year + " ...");
      aeData.readYear(year);
      //Get the data as a Vector of AE objects.
      Vector vec = aeData.getVector();
      //Store in database.
      System.out.print("Writing to database ... ");
      //Store in database if the vector contain data.
      if (!vec.isEmpty()) {
        try {
          ObjectTimeSeries ts = new ObjectTimeSeries(vec);
          Database db = Database.getInstance("AEDB");
          db.store(ts);
          System.out.println("Done!");
        }
        catch (IllegalTimeSeriesException e) {
          System.err.println(e);
        }
      }
      else {
        System.out.println("Failed! No data to store.");
      }
    }
  }
  
  class AEDataOneYear {
    //The year.
    int year;
    //The data.
    // Month, day, hour, and index type.
    double[][][][] oneYearData = new double[12][31][24][4];
    //The number of days in each month.
    //The index types.
    int AE = 0;
    int AL = 1;
    int AO = 2;
    int AU = 3;
    //Calendar object for time operations.
    Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
    
    public AEDataOneYear() {
    }
    
    public void readYear(int year) {
      this.year = year;
      //Reset all data.
      for (int m=0;m<oneYearData.length;m++) {
        for (int d=0;d<oneYearData[m].length;d++) {
          for (int h=0;h<oneYearData[m][d].length;h++) {
            for (int i=0;i<oneYearData[m][d][h].length;i++) {
              oneYearData[m][d][h][i] = Double.NaN;
            }
          }
        }
      }
      //The two digit year.
      int yy;
      if (year<2000) {
        yy = year - 1900;
      }
      else {
        yy = year -2000;
      }
      try {
        //Open the file.
          String inputFileName = inputDir + "ae" + yy + "hv";
        BufferedReader bufReader = new BufferedReader(new FileReader(inputFileName));
        //Read all lines in the file.
        String textLine;
        while ((textLine=bufReader.readLine())!=null) {
          int m = this.getMonth(textLine)-1;
          int d = this.getDay(textLine)-1;
          int i = this.getIndex(textLine);
          for (int h=0;h<24;h++) {
            oneYearData[m][d][h][i] = this.getDataAtHour(textLine,h);
          }
        }
        //Close the buffer.
        bufReader.close();
      }
      catch (IOException e) {
        System.err.println(e);
//        System.exit(0);
      }
    }
    
    public Vector getVector() {
      Vector vec = new Vector(366*24);
      this.cal.clear();
      this.cal.set(Calendar.YEAR,this.year);
      while (cal.get(Calendar.YEAR) == this.year) {
        int m = cal.get(Calendar.MONTH);
        int d = cal.get(Calendar.DATE)-1;
        int h = cal.get(Calendar.HOUR_OF_DAY);
        //This check is a quick-and-dirty solution.
        //The input data ends at 1988-06-30 so the algorithm
        //will just fill with data gaps for the rest of 1988.
        //This check will not put any data into the vector
        //that comes after 1988-06 (Java month 5 = Real month 6).
        if (this.year==1988 & m>5) {
        }
        else {
            AE ae = new AE(cal.getTime(),oneYearData[m][d][h]);
            vec.addElement(ae);
        }
        cal.add(Calendar.HOUR_OF_DAY,1);
      }
      vec.trimToSize();
      return vec;
    }
    
    private int getMonth(String textLine) {
      int month = 0;
      try {
        month = Integer.parseInt(textLine.substring(5,7));
      }
      catch (NumberFormatException e) {
        System.err.println(e);
        System.exit(0);
      }
      return month;
    }
    
    private int getDay(String textLine) {
      int day = 0;
      try {
        day = Integer.parseInt(textLine.substring(8,10));
      }
      catch (NumberFormatException e) {
        System.err.println(e);
        System.exit(0);
      }
      return day;
    }
    
    private int getIndex(String textLine) {
      switch (textLine.charAt(7)) {
        case 'E':
          return this.AE;
        case 'L':
          return this.AL;
        case 'O':
          return this.AO;
        case 'U':
          return this.AU;
        default:
          return -1;
      }
    }
    
    private double getDataAtHour(String textLine, int hour) {
      int start = 20 + 4*hour;
      int end = 24 + 4*hour;
      double base = 0.0;
      double data = 0.0;
      //Try to read the base value, it does not always exist.
      try {
        base = Double.parseDouble(textLine.substring(16,20));
      }
      catch (NumberFormatException e) {
        //If it is not the empty string error then exit.
        if (!e.getMessage().equalsIgnoreCase("empty String")) {
          System.err.println("Error reading the base value.");
          System.err.println(e);
          System.exit(0);
        }
      }
      //Try to read the data value, it should always exist.
      try {
        data = Double.parseDouble(textLine.substring(start,end));
      }
      catch (NumberFormatException e) {
          System.err.println("Error reading the data value.");
          if (e.getMessage().equalsIgnoreCase("empty String")) {
            System.err.println("Missing data value.");
            System.err.println(e);
            System.err.println("Replacing with NaN.");
            data = Double.NaN;
          }
          else {
            System.err.println(e);
            System.exit(0);
          }
      }
      //Is it a data gap?
      if (data==9999.0) {
        data = Double.NaN;
      }
      return 100.0*base + data;
    }
    
  }
      
  /**
  * @param args the command line arguments
  */
  public static void main (String args[]) {
      if (args.length != 3) {
          System.err.println("Usage: BuildAE <input directory> <start year> <end year>");
          System.exit(0);
      }
    //Create the build AE class.
      int sy = Integer.parseInt(args[1]);
      int ey = Integer.parseInt(args[2]);
    BuildAE ae = new BuildAE(args[0],sy,ey);
    //Run the program.
    try {
        ae.run();
    }
    catch (Exception e) {
        System.err.println(e);
    }
  }

}