package saaps.dbt;

import java.io.*;
import java.util.*;
import java.text.*;

import saaps.util.*;

/**
 * ACEDB extends the Database class. It contains
 * the read and write methods that are used in
 * Database. The field is a vector of ACE objects.
 *
 *@authors Peter Wintoft, Leif Kalla
 *@version 1.1 1999-12-21 (Use new ACEmag class)
 *@version 1.0 990913
 *@see ACEmag.html
 */
public class DstDB extends Database {

  private double dataGapValueDefinition = -999.9;

  // Static Fields used to initiate the MetaData
  private static String name = "DstDB";
  private static String description = "Hourly Dst index.";
  private static String longDescription = "The final Dst index covers the period 1957 to 1998." +
  "The years 1957-1998 contain the final Dst, while the years 1999 and 2000 contain the preliminary Dst";

  private static Calendar st=Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
  private static Calendar et=Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
  private static boolean ct=true;
  private static long step=3600000L; //One hour timestep 
  private static String[] fieldNames = { "Dst" , "Status" };
  private static String[] fieldUnits = { "nT" , ""};
  private static String[] fieldTypes = { "Double" , "Double"};

  // Fields.
  private Dst dst;
  private Vector dstVec;

  // Constructors.
  public DstDB() {
    //
    dataGapValue=dataGapValueDefinition;
  }
  
  public DstDB(int i) {
    //
    this.dataGapValue=this.dataGapValueDefinition;
  }

  /**
   * When a new class is added to the database this
   * constructor should be used which includes the
   * information that goes into the meta data file.
   *
   * The arguments are: Vector of ACE objects, the class name,
   * a short description, continuos time series (false/true), and
   * if a continous time series the time step.
   */
  public DstDB(Vector dstVec) throws Exception {
    this.dataGapValue=this.dataGapValueDefinition;
    this.dstVec = dstVec;
    this.raf = new RandomAccessFile(baseDir + this.name + ".data","rw");
    // Then get its metadata
    try { // To load the meta data for this database from file
      this.meta = new MetaData(this.name);
      this.isSet=true;
    }
    catch (IOException e) {
      System.err.println(e);
      System.out.println("Here we should handle the situation when no DB exists");
      try {
        this.meta = setDefaultMeta(((DatabaseObject) (this.dstVec.elementAt(0))).getDate(),
                              ((DatabaseObject) (this.dstVec.elementAt(this.dstVec.size()-1))).getDate()); //Takes static fields from XDB
        this.isSet=true;
      }
      catch (IOException ee) {
        System.err.println(ee);
        ee.printStackTrace();
      }
    }
    catch (ClassNotFoundException e) {
      System.err.println(e);
      e.printStackTrace();
    }

    store((Vector) this.dstVec); //Need to type cast so that store(Object obj) isn't used
  }

  // Methods.
  public Object read() throws IOException {
    this.dst = new Dst(this.raf.readLong(),this.raf.readDouble(),this.raf.readDouble());
    return (Object) this.dst;
  }

  public void write(Object obj) throws IOException {
    this.dst = (Dst) obj;
    this.raf.writeLong(this.dst.getDate().getTime());
    this.raf.writeDouble(this.dst.getDst());
    this.raf.writeDouble(this.dst.getStatus());
  }

  // Implemented for Database abstract getDate
  public Date getDate(Object obj) {
    return ((Dst) obj).getDate();
  }

  /** Creates a metadata file for this class with the default values set
   *to the static values defined above. OBS!, check calendar ....
   */
  protected MetaData setDefaultMeta(Date sd, Date ed) throws IOException {

    this.st.setTime(sd);
    this.st.clear(st.MILLISECOND);
    this.et.setTime(ed);
    this.et.clear(et.MILLISECOND);

    MetaData m = new MetaData(name,description, longDescription,  st, et, ct, step,fieldNames,fieldUnits,fieldTypes);
    return m;
  }

  /**
   * Returns an "empty" metadata object. All fields but Start and end time
   * has valid values (as set in the respective database class).
   * The times will be set to that what is supplied.
   * OBS! This is not stored in the MetaData file !!
   */
  protected MetaData getEmptyMetaData(Date sd, Date ed) {
    MetaData m = new MetaData();
    this.st.clear(st.MILLISECOND);
    this.st.setTime(sd);
    this.et.clear(et.MILLISECOND);
    this.et.setTime(ed);
    m.setMetaData(name,description, longDescription,  st, et, ct, step, fieldNames,fieldUnits,fieldTypes );
    return m;
  }

    /**
     * Just a "placeholder" so that it will be possible to compile
     * this file. This method will not be used ! 
     */
    public Vector getLatestDatabaseObjects() {
	return new Vector(1);
    }
}
