package saaps.dbt;

import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;

import saaps.dbt.*;
import saaps.util.ObjectTimeSeries;
import saaps.util.IllegalTimeSeriesException;

/**
 * DstDownloader downloads Dst data from worlddatacenter in Japan
 * and stores it in the SAAPS database.
 *
 * NEAR_REAL_TIME
 * http://swdcdb.kugi.kyoto-u.ac.jp:80/dstdir/dst1/q/Dstqthism.html
 * http://swdcdb.kugi.kyoto-u.ac.jp:80/dstdir/dst1/q/Dstqlastm.html
 * http://swdcdb.kugi.kyoto-u.ac.jp:80/dstdir/dst1/q/Dstqlastm2.html
 *
 * PROVISIONAL
 * http://swdcdb.kugi.kyoto-u.ac.jp/dstdir/dst1/p/dstprov200001.html
 * http://swdcdb.kugi.kyoto-u.ac.jp/dstdir/dst1/p/dstprov200002.html
 * and so on .... until end of 2000
 *
 * FINAL
 * http://swdcdb.kugi.kyoto-u.ac.jp/dstdir/dst1/f/dstfinal195701.html
 * http://swdcdb.kugi.kyoto-u.ac.jp/dstdir/dst1/f/dstfinal195702.html 
 * and so on until 1999
 *
 *@author Leif Kalla
 */

public class DstDownloader {
    
    private static double dataGapValue = 9999;
    private DateFormat df;
    private Hashtable htbl;
    
    public static String firstMatcher="for Geomagnetism"; //WDC-C2 
    public Hashtable monthHTBL;
    
    protected static String[] URLlist = { "http://swdcdb.kugi.kyoto-u.ac.jp:80/dstdir/dst1/q/Dstqthism.html",
    "http://swdcdb.kugi.kyoto-u.ac.jp:80/dstdir/dst1/q/Dstqlastm.html",
    "http://swdcdb.kugi.kyoto-u.ac.jp:80/dstdir/dst1/q/Dstqlastm2.html"
    };
    
    private static String provisionalTemplate = "http://swdcdb.kugi.kyoto-u.ac.jp/dstdir/dst1/p/dstprov";
    private static String finalTemplate = "http://swdcdb.kugi.kyoto-u.ac.jp/dstdir/dst1/f/dstfinal";
    
    private static String[] dataTypeStr = {"FINAL","PROVISIONAL"," NEAR REAL TIME "};
    private static String[] theMonths = {
        "JANUARY",
        "FEBRUARY",
        "MARCH",
        "APRIL",
        "MAY",
        "JUNE",
        "JULY",
        "AUGUST",
        "SEPTEMBER",
        "OCTOBER",
        "NOVEMBER",
        "DECEMBER"
    };

    private static int dataStringLength = 101;
    private static int dataLentgth = 4;
    private static int startPos = 3;
    private static int dataPositions[] = {
        DstDownloader.startPos,
        DstDownloader.startPos+4*1,
        DstDownloader.startPos+4*2,
        DstDownloader.startPos+4*3,
        DstDownloader.startPos+4*4,
        DstDownloader.startPos+4*5,
        DstDownloader.startPos+4*6,
        DstDownloader.startPos+4*7,
        DstDownloader.startPos+4*8+1,
        DstDownloader.startPos+4*9+1,
        DstDownloader.startPos+4*10+1,
        DstDownloader.startPos+4*11+1,
        DstDownloader.startPos+4*12+1,
        DstDownloader.startPos+4*13+1,
        DstDownloader.startPos+4*14+1,
        DstDownloader.startPos+4*15+1,
        DstDownloader.startPos+4*16+2,
        DstDownloader.startPos+4*17+2,
        DstDownloader.startPos+4*18+2,
        DstDownloader.startPos+4*19+2,
        DstDownloader.startPos+4*20+2,
        DstDownloader.startPos+4*21+2,
        DstDownloader.startPos+4*22+2,
        DstDownloader.startPos+4*23+2
    };
        
    public DstDownloader() {
        df = DateFormat.getDateInstance(DateFormat.SHORT,new Locale("sv","SE"));
        df.setTimeZone(TimeZone.getTimeZone("UTC"));
        
        /* Map the data status string to integers 1 - 3, were 1 is FINAL 2 is PROVISIONAL , 3 is NEAR REAL TIME */
        htbl = new Hashtable();
        htbl.put(dataTypeStr[0],new Integer(1));
        htbl.put(dataTypeStr[1],new Integer(2));
        htbl.put(dataTypeStr[2],new Integer(3));

        monthHTBL = new Hashtable();
        for(int i=0; i<12;i++) {
            monthHTBL.put(DstDownloader.theMonths[i],new Integer(i+1));
        }
    }

    /**
     * Constructs the PROVISIONAL Dst URL from the template string and input arguments.
     * @param year 
     * @param month
     * @return The String that represents the PROVISIONAL data location
     */
    public static String constructProvisionalURL(String year,String month) {
        return new String(DstDownloader.provisionalTemplate+year+month+".html");
    }

    /**
     * Constructs the FINAL Dst URL from the template string and input arguments.
     * @param year 
     * @param month
     * @return The String that represents the FINAL data location
     */
    public static String constructFinalURL(String year,String month) {
        return new String(DstDownloader.finalTemplate+year+month+".html");
    }
    
    /**
     * Create an array of Dst objects from the string and the Calendar (year, month)
     * @param txt the text line from the world datacenter
     * @param cal the year and month are stored here
     * @return array of Dst objects
     * @exception DstParseException when the textline coulden't be parsed
     */
    public static void parseDataString(Vector data,String txt,Calendar cal,int status) throws DstParseException {
        if(txt.length() != DstDownloader.dataStringLength) {
            throw new DstParseException("Not correct length for data string!");
        }
        /* THE DATE */
        int day;
        if(txt.charAt(0) == ' ')
            day = Integer.parseInt(txt.substring(1,2));
        else
            day = Integer.parseInt(txt.substring(0,2));

        cal.set(Calendar.DAY_OF_MONTH,day);
        /* go through all hours of the day */
        for(int i=0;i<24;i++) {
            cal.set(Calendar.HOUR_OF_DAY,i);
            double tmp = Double.parseDouble(txt.substring(DstDownloader.dataPositions[i],DstDownloader.dataPositions[i]+DstDownloader.dataLentgth));
            if ( tmp == DstDownloader.dataGapValue )
                tmp = Double.NaN;
            data.add(new Dst(cal.getTime(),tmp,status));
        }
    }
    
    public void getData(String urlStr) {
        Date d;
        try {
            URL url = new URL(urlStr);
            URLConnection urlConn = url.openConnection();
            // Input stream.
            BufferedReader streamIn = new BufferedReader(
            new InputStreamReader(
            urlConn.getInputStream()));
            //System.out.println("Got input stream");
            String textLine = "";
            int month=1;
            int year=1;
            int status;
            
            /* Find the datatype position */
            textLine = streamIn.readLine();
            while ((textLine = streamIn.readLine()) != null) {
                if(textLine.indexOf('(') == -1 || textLine.indexOf(')') == -1) 
                    continue;
                else
                    break;
            }

            /* Find datatype pos */
            if(textLine != null) {
                String type = textLine.substring(textLine.indexOf('(')+1,textLine.lastIndexOf(')'));
                  /* Check if it's a valid key */
                if(htbl.get(type) == null) {
                    streamIn.close();
                    throw new DatabaseException("Wrong status for hash:"+type);
                }
                status = ((Integer) htbl.get(type)).intValue();
            }
            else {
                streamIn.close();
                throw new DatabaseException("Cannot locate data status position");
            }

            /* Find Month and year */
            while((textLine = streamIn.readLine()) != null) {
                if(!textLine.equals(""))
                    break;
            }
            
            if(textLine != null) {
                StringTokenizer tokens = new StringTokenizer(textLine);
                String value = null;

                int numTok = tokens.countTokens();
                if(numTok != 2) {
                    streamIn.close();
                    throw new DatabaseException("Not correct number of tokens in Month/Year");
                }
                String tmp = tokens.nextToken();
                if(monthHTBL.get(tmp) == null) {
                    streamIn.close();
                    throw new DatabaseException("Wrong month:"+tmp);
                }
                month = ((Integer) monthHTBL.get(tmp)).intValue();
                tmp = tokens.nextToken();
                year = Integer.parseInt(tmp);
            }
            else {
                streamIn.close();
                throw new DatabaseException("Cannot get month and year info.");
            }
            System.out.println("y:"+year+" m:"+month);

            Calendar cal = Calendar.getInstance();
            cal.setTimeZone(TimeZone.getTimeZone("UTC"));
            cal.set(year,month-1,1,0,0,0);
            cal.clear(cal.MILLISECOND);
            
            /* search for first data row */
            textLine = "HEJ";
            while ( textLine != null && !textLine.startsWith(" 1")) {
                textLine = streamIn.readLine();
            }
            if(textLine == null) {
                streamIn.close();
                throw new DatabaseException("No data found");
            }
            /* OK continue with data parsing */
            Vector data = new Vector(31*24);
            while(textLine != null) {
                try {
                    if(textLine.equals("")) {
                        textLine = streamIn.readLine();
                        continue;
                    }
                    DstDownloader.parseDataString(data,textLine,cal,status);
                    textLine = streamIn.readLine();
                }
                catch (DstParseException ee) {
                 /* When this happen we were not able to parse the supplied string */
                 /* Assume we are ready and put the data into the DB */
                    
                    /* remove last NaNs */
                    for(int i=data.size()-1;i>-1;i--) {
                        if(Double.isNaN(((DatabaseObject) data.elementAt(i)).getData(0)))
                            data.remove(i);
                        else
                            break;
                    }
                    data.trimToSize();
                    ObjectTimeSeries ts = new ObjectTimeSeries(data);
                    
                    Database db = Database.getInstance("DstDB");
                    db.store(ts);                    
                    break;
                }
            }
            streamIn.close();            
        }
        catch (Exception e) {
            System.err.println("error: " + e);
            e.printStackTrace();
        }        
    }
    
    public static void main(String[] args) {
        int type = 0; //Default is latest data
        
        if(args.length>0) {
            type = Integer.parseInt(args[0]);
        }
        DstDownloader ldst = new DstDownloader();
        ldst.getData(DstDownloader.URLlist[type]);
    }
}
