package saaps.dbt;

import java.text.DateFormat;
import java.text.NumberFormat;

import java.util.Calendar;
import java.util.Date;
import java.util.Locale;
import java.util.TimeZone;
import java.io.Serializable;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import java.util.Vector;

/**
* A class for GOES-08 X-ray data.
*
* The class contains the date, short (0.05-0.4 nm) and
* long (0.1-0.8 nm) X-rays.
* It also contain the longitude position of the satellite.
*
* @author Peter Wintoft
* @version 1.0 1999-12-22
*/
public class GOES08xray extends DatabaseObject implements Serializable {
  // Fields
  private static int numberOfIndexes = 3;
  private Date d;
  private double shortXR, longXR, longitude;

  // Constructors
  public GOES08xray()
  {
  }

  /**
   * This constructor is used if you want an "empty" object with this date.
   */
  public GOES08xray(Date D) {
    d = D;
    shortXR = Double.NaN;
    longXR = Double.NaN;
    longitude = Double.NaN;
  }

  public GOES08xray(Date D, double SXR, double LXR, double LG)
  {
    d = D;
    shortXR = SXR;
    longXR = LXR;
    longitude = LG;
  }
  public GOES08xray(long L, double SXR, double LXR, double LG)
  {
    this(new Date(L),SXR,LXR,LG);
  }

  public GOES08xray(Date D, double[] data) {
    this(D,data[0],data[1],data[2]);
  }

  // Methods
  public double getShortXR()
  {
    return shortXR;
  }
  public double getLongXR()
  {
    return longXR;
  }
  public double getLongitude()
  {
    return longitude;
  }

  //Implemented for DatabaseObject
  public Date getDate()
  {
    return d;
  }

  // This method is the generic form of accessing data.
  public Vector getData()
  {
    Vector vec = new Vector(numberOfIndexes);
    vec.addElement(new Double(shortXR));
    vec.addElement(new Double(longXR));
    vec.addElement(new Double(longitude));
    return vec;
  }

  /**
   * This method returns the size of the data that is to be written to file
   */
  public long getDataSize() throws IOException {

    ByteArrayOutputStream baos = new ByteArrayOutputStream();  
    DataOutputStream    dos  = new DataOutputStream(baos);  
    dos.writeLong(d.getTime());
    dos.writeDouble(shortXR);
    dos.writeDouble(longXR);
    dos.writeDouble(longitude);

    dos.flush();  
    long size = baos.size();  // now you know how large ....
    return size;
  }

  public void setDate(Date D)
  {
    d = D;
  }
  public void setDate(long L)
  {
    d = new Date(L);
  }
  public void setShortXR(double SXR)
  {
    shortXR = SXR;
  }
  public void setLongXR(double LXR)
  {
    longXR = LXR;
  }
  public void setLongitude(double LG)
  {
    longitude = LG;
  }

  //Implemented for DatabaseObject getDate
  public int getNumberOfDataIndexes() {
    return (numberOfIndexes);
  }

  //Implemented for DatabaseObject
  public double getData(int index) {
    //Date is not considered a data value
    switch (index) {
    case 0 :
      return(shortXR);
    case 1 :
      return(longXR);
    case 2 :
      return(longitude);
    default :
      System.out.println("Index out of bounds for GOES08xray");
      return (Double.NaN);
    }
  }

  public String toString() {
//     // Set the date format.
//     DateFormat df = DateFormat.getDateTimeInstance(DateFormat.DEFAULT,DateFormat.DEFAULT,new Locale("sv","SE")); //en UK
//     df.setTimeZone(TimeZone.getTimeZone("UTC"));

    // Find the selected dates.
    Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("UTC"));
    cal.clear(cal.MILLISECOND);
    cal.setTime(d);

    NumberFormat nf = NumberFormat.getInstance();
    nf.setMinimumIntegerDigits(2);

    return ( cal.get(Calendar.YEAR) + " " + nf.format(cal.get(Calendar.MONTH)+1) + " " + nf.format(cal.get(Calendar.DAY_OF_MONTH)) + " " +
             nf.format(cal.get(Calendar.HOUR_OF_DAY)) + " " + nf.format(cal.get(Calendar.MINUTE)) + " " + nf.format(cal.get(Calendar.SECOND)) + " " +
             Double.toString(shortXR) + " " + Double.toString(longXR) + " " + Double.toString(longitude) );

  }

}
