package saaps.dbt;

import java.io.*;
import java.util.*;
import java.net.*;
import java.text.*;

import saaps.util.ObjectTimeSeries;
import saaps.util.IllegalTimeSeriesException;

/**
 * KpDB extends the Database class. It contains
 * the read and write methods that are used in
 * Database. The field is a vector of Kp objects.
 *
 *@author Peter Wintoft, Leif Kalla
 *@version 1.0 990718
 */
public class KpDB extends Database {

  // Absolute path to the database.
  //   private String baseDir = "/space/saaps/Database/";

  private static final String dataProvider="ftp://ftp.sel.noaa.gov/pub/latest/DGD.txt";
  private static final double dataGapValueDefinition = -1.0;


  // Static Fields used to initiate the MetaData
  private static String name="KpDB";
  private static String description = "Kp index";
  private static String longDescription = "The Kp index is a 3-hour index that "+
    "summarizes the global planetary geomagnetic disturbance level. " +
    "The index range from 0 to 9, where 0 is quiet and 9 severe storm. " +
    "Currently, there are four versions of the Kp indicated by the Status field." +
    "1=final Kp, 2=estimated Kp from SEC, 3=nowcasted Kp from Lund, 4=3-hour forecasted Kp from Lund";
  private static Calendar st=Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
  private static Calendar et=Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
  private static boolean ct=true; //This data is continous
  private static long step=10800000; //Every three hour we have a new value
  private static String[] fieldNames = { "Kp" , "Status"};
  private static String[] fieldUnits = { "" , ""};
  private static String[] fieldTypes = { "Double" , "Double"};

  //
  private Kp kp;
  private Vector kpv;


  // Constructors.
  //   public KpDB() {
  //       
  //       kp = new Kp(1999,i);
  //     }

  //Constructors, every time you create an instance of this database
  //you should also get its associated meta data.
  //If matadata is unavailable , try to create a default metadata file.
  public KpDB() {
    
    dataGapValue=dataGapValueDefinition;
  }

  public KpDB(double i) {
    
    dataGapValue=dataGapValueDefinition;
    kp = new Kp(1999,i);
    //      store(kpv);
  }
  /**
   * When a new class is added to the database this
   * constructor should be used which includes the
   * information that goes into the meta data file.
   *
   * The arguments are: Vector of Kp objects, the class name,
   * a short description, continuos time series (false/true), and
   * if a continous time series the time step.
   */
  public KpDB(Vector KPV) throws Exception {
    //, String des, boolean ct, long step) {
    //       
    
    dataGapValue=dataGapValueDefinition;

    raf = new RandomAccessFile(baseDir + "KpDB.data","rw");
    kpv = KPV;

    // Then get its metadata
    try { // To load the meta data for this database from file
      meta = new MetaData("KpDB");
      isSet=true;
    }
    catch (IOException e) {
      System.err.println(e);
      System.out.println("Here we should handle the situation when no DB exists");
      try {
	meta = setDefaultMeta(((DatabaseObject) (kpv.elementAt(0))).getDate(),((DatabaseObject) (kpv.elementAt(kpv.size()-1))).getDate()); //Takes static fields from XDB

	isSet=true;
      }
      catch (IOException ee) {
	System.err.println(ee);
	ee.printStackTrace();
      }
    }
    catch (ClassNotFoundException e) {
      System.err.println(e);
      e.printStackTrace();
    }

    //And now store the data
    store((Vector) kpv);
  }

  //  Implemented for Database abstract read
  public Object read() throws IOException {
    kp = new Kp(raf.readLong(),raf.readDouble(),raf.readDouble());
    return (Object) kp;
  }

  // Implemented for Database abstract write
  public void write(Object obj) throws IOException  {
    kp = (Kp) obj;
    raf.writeLong(kp.getDate().getTime());
    raf.writeDouble(kp.getKp());
    raf.writeDouble(kp.getStatus());
  }

  // Implemented for Database abstract getDate
  public Date getDate(Object obj) {
    return ((Kp) obj).getDate();
  }

  /** This method is implemented for Database abstract metod.
   * It retrieves the latest data from the dataprovider and returns the vector of Database objects.
   */
  public Vector getLatestDatabaseObjects() {

    String dateStr =  null;
    int h;
    DateFormat df;
    Date d;

    // Create a Vector of Kp objects.
    Vector kpv = new Vector(8*30);

    try {
      // Open connection
      URL url = new URL(dataProvider);
      URLConnection urlConn = url.openConnection();
      // Input stream.
      BufferedReader streamIn = new BufferedReader(new InputStreamReader(urlConn.getInputStream()));

      //Setup dateformat
      df = DateFormat.getDateInstance(DateFormat.SHORT,new Locale("sv","SE"));
      df.setTimeZone(TimeZone.getTimeZone("UTC"));
      Calendar cal = Calendar.getInstance();
      cal.setTimeZone(df.getTimeZone());

      String textLine = "";
      textLine = streamIn.readLine();
      // Go through the file for all lines starting with
      // "#", ":", or " ".
      while (!textLine.startsWith("1") && !textLine.startsWith("2")) {
	textLine = streamIn.readLine();
      }

      do {
	// Read the date using - as the field separator.
	dateStr = textLine.substring(0,4) + "-" +
	  textLine.substring(5,7) + "-" +
	  textLine.substring(8,10);
	d = df.parse(dateStr);
	// Assume that observations are at 00, 03, 06, ...
	for (int i=0;i<8;i++) {
	  h=3*i;
	  cal.setTime(d);
	  cal.add(Calendar.HOUR,h);
	  double tmpKp = Double.valueOf(textLine.substring(63+2*i,65+2*i).trim()).doubleValue();
	  if (tmpKp == dataGapValue ) {
	    tmpKp = Double.NaN;
	  }
	  Kp kpTmp = new Kp(cal.getTime().getTime(),tmpKp);
	  kpv.addElement(kpTmp);
	  //System.out.println("get: " + kpTmp.getDate());
	  // Integer.valueOf(textLine.substring(63+2*i,65+2*i).trim()).intValue()));
	}
      }
      while ((textLine = streamIn.readLine()) != null);

      // Trim the vector.
      kpv.trimToSize();
      // Close input stream.
      streamIn.close();
    }
    catch (IOException e) {
      System.err.println("URL Connection error: " + e);
    }
    catch (NumberFormatException e) {
      System.err.println("Number format error: " + e);
    }
    catch (ParseException e) {
      System.err.println("Parse error: " + e);
    }

    return kpv;
  }

  /** 
   * This method is used for importing ASCII data to our database.
   * The format of the data is as output from ExportAscii application.
   * It will fill a buffer of databaseobjects and then try to store it,
   * read again and store until it is finished or an exception occures.
   */
  protected void importAsciiOLD(String inputfile) throws Exception {
    final int MAXSIZE=1048576; // bytes , 1Mb 
    //final int MAXSIZE=1024; // bytes ,

    Date inputDateTime;
    String tmpVal;
    double tmpKp;


    // Create a Vector of Kp objects.
    Vector kpv = new Vector(8*30);

    try {
      //set default maximum number of objects
      int maxObjs = 100;
      if(meta.getCont()) { //Calculate if we have continous data
	//get size of kp object
	long objSize = (new Kp(new Date())).getObjectSize();
	maxObjs = (int) (MAXSIZE/objSize);
      }

      int cnt=0;

      // Input stream.
      BufferedReader streamIn = new BufferedReader(new FileReader(inputfile));
      String textLine = "";
      textLine = streamIn.readLine();
      // Go through the file for all lines starting with
      // "#", ":", or " ".
      while (!textLine.startsWith("1") && !textLine.startsWith("2")) {
	textLine = streamIn.readLine();
      }

      while(textLine != null) {
	while ((textLine != null) && ( cnt < maxObjs) ) {
	  cnt++;
	  // Read the date using - as the field separator.
	  String dateStr = textLine.substring(0,4) + "." +
	    textLine.substring(5,7) + "." +
	    textLine.substring(8,10) + " " + textLine.substring(11,13) + ":" + textLine.substring(14,16) + ":" + textLine.substring(17); //Last item to end of string

	  SimpleDateFormat formatter = (SimpleDateFormat) SimpleDateFormat.getDateTimeInstance(0,0,new Locale("sv","SE"));

	  formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
	  formatter.applyPattern("yyyy.MM.dd HH:mm:ss");
     
	  // Parse the string back into a Date.
	  ParsePosition pos = new ParsePosition(0);
	  inputDateTime = formatter.parse(dateStr, pos);
	
	// Assume that observations are at 00, 03, 06, ...
	  Calendar cal = Calendar.getInstance();
	  cal.setTimeZone(formatter.getTimeZone());
	  cal.setTime(inputDateTime);
	  //The data value
	  tmpVal = textLine.substring(20).trim();
	  if(tmpVal.equals(new String("NaN"))) {
	    tmpKp = Double.NaN;
	  }
	  else {
	    tmpKp = Double.valueOf(tmpVal).doubleValue();
	  }
	  if (tmpKp == dataGapValue ) {
	    tmpKp = Double.NaN;
	  }

	  //System.out.println("Date/tome:" + formatter.format(inputDateTime) + " " + tmpKp);

	  kpv.addElement(new Kp(cal.getTime().getTime(),tmpKp));
	  //Get the next line
	  textLine = streamIn.readLine();	
	}
	//Store our buffer of objects
	kpv.trimToSize();
	// Store the data in the database
	ObjectTimeSeries ts = new ObjectTimeSeries(kpv);

	if(!isSet())
	  System.out.println("MetaData for Kp is unavailable \nWill try to create it.");
	System.out.println("Storing:" + cnt);
	store(ts);

	//Reset the vector and counter
	//kpv.removeAllElements();
	kpv = new Vector(8*30);
	cnt=0;
      }
      streamIn.close();
    }
    catch (IOException e) {
      System.err.println("IO error: " + e);
    }
    catch (NumberFormatException e) {
      System.err.println("Number format error: " + e);
    }
    catch (IllegalTimeSeriesException e) {
      System.err.println(e);
    }

  }

  /** Creates a metadata file for this class with the default values set 
   *to the static values defined above. OBS!, check calendar ....
   */
  protected MetaData setDefaultMeta(Date sd, Date ed) throws IOException {

    st.clear(st.MILLISECOND);
    st.setTime(sd);
    et.clear(et.MILLISECOND);
    et.setTime(ed);
    MetaData m = new MetaData(name,description, longDescription,  st, et, ct, step, fieldNames, fieldUnits, fieldTypes );
    return m;
  }

  /**
   * Returns an "empty" metadata object. All fields but Start and end time
   * has valid values (as set in the respective database class).
   * The times will be set to that what is supplied.
   * OBS! This is not stored in the MetaData file !!
   */
  protected MetaData getEmptyMetaData(Date sd, Date ed) {
    MetaData m = new MetaData();
    st.clear(st.MILLISECOND);
    st.setTime(sd);
    et.clear(et.MILLISECOND);
    et.setTime(ed);
    m.setMetaData(name,description,longDescription,  st, et, ct, step, fieldNames, fieldUnits, fieldTypes );
    return m;
  }

  public void myHello() {
    System.out.println("My KpDB Hello");
  }

  public void yourHello() {
    System.out.println("Your KpDB Hello: " + meta.getDescr());
  }

}
