package saaps.dbt;

import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;

import saaps.util.*;

/**
* LatestGOES08part downloads the latest GOES08 particle data and position.
* The fields are:
* P>1 MeV, P>5 MeVB, P>10 MeV, P>30 MeV, P>50 MeV, P>100 MeV, E>0.6 MeV,
* E>2 MeV, E>4 MeV, and longitude.
* FTP address: ftp://ftp.sec.noaa.gov/pub/lists/particle/G8part_5m.txt
*
*@author Peter Wintoft
*@version 1.0 2000-02-22
*/
public class LatestGOES08part extends Object
{
  private static double dataGapValue = -1.00e+05;
  public static void main(String[] args)
  {
    String ftpSrcStr="ftp://ftp.sec.noaa.gov/pub/lists/particle/G8part_5m.txt";
    DateFormat df;
    Date d;
    Vector goesv = null;
    double longitude = Double.NaN;
    try
    {
      // Open connection
      URL url = new URL(ftpSrcStr);
      URLConnection urlConn = url.openConnection();
      // Input stream.
      BufferedReader streamIn = new BufferedReader(
                                  new InputStreamReader(
                                    urlConn.getInputStream()));
      String textLine = "";
      textLine = streamIn.readLine();
      // Go through the file for all lines starting with
      // "#", ":".
//      while (!textLine.startsWith("1") && !textLine.startsWith("2"))
      while (textLine.charAt(0)=='#' || textLine.charAt(0)==':')
      {
//        System.out.println("Skipping: " + textLine);
        // Get the longitude value.
        if (textLine.startsWith("# Location:"))
          longitude = Double.valueOf(textLine.substring(13,16)).doubleValue();
        textLine = streamIn.readLine();
      }
      // Create a Vector of GOES08part objects.
      goesv = new Vector(24);
      while ((textLine = streamIn.readLine()) != null)
      {
//        System.out.println("Reading: " + textLine);
        // Tokenize the string.
        StringTokenizer st = new StringTokenizer(textLine);
        // Create the date string.
        String dateStr = st.nextToken() + "-" +    // Year
                          st.nextToken() + "-" +  // Month
                          st.nextToken() + " ";    // Day
        String timeStr = st.nextToken();          // Hour and minute
        // Insert a ":" between hour and minute.
        timeStr = timeStr.substring(0,2) + ":" + timeStr.substring(2,4);
        dateStr = dateStr + timeStr;
        // Read the date using - as the field separator.
//        String dateStr = textLine.substring(0,4) + "-" +
//                           textLine.substring(5,7) + "-" +
//                           textLine.substring(8,10) + " " +
//                           textLine.substring(11,13) + ":" +
//                           textLine.substring(13,15);
        // Use UTC time and get the Date object.
        df = DateFormat.getDateTimeInstance(DateFormat.SHORT,DateFormat.SHORT,new Locale("sv","SE"));
        df.setTimeZone(TimeZone.getTimeZone("UTC"));
        d = df.parse(dateStr);
        // Skip over until P>1 field.
        st.nextToken(); //Julian day.
        st.nextToken(); //Seconds of the day.
        // Read P>1, P>5, P>10, P>30, P>50, P>100, E>0.6, E>2, E>4, and longitude.
        double p1 = Double.valueOf(st.nextToken()).doubleValue();
        double p5 = Double.valueOf(st.nextToken()).doubleValue();
        double p10 = Double.valueOf(st.nextToken()).doubleValue();
        double p30 = Double.valueOf(st.nextToken()).doubleValue();
        double p50 = Double.valueOf(st.nextToken()).doubleValue();
        double p100 = Double.valueOf(st.nextToken()).doubleValue();
        double e06 = Double.valueOf(st.nextToken()).doubleValue();
        double e2 = Double.valueOf(st.nextToken()).doubleValue();
        double e4 = Double.valueOf(st.nextToken()).doubleValue();
        // Replace data gaps with NaNs.
        if (p1 == dataGapValue)
          p1 = Double.NaN;
        if (p5 == dataGapValue)
          p5 = Double.NaN;
        if (p10 == dataGapValue)
          p10 = Double.NaN;
        if (p30 == dataGapValue)
          p30 = Double.NaN;
        if (p50 == dataGapValue)
          p50 = Double.NaN;
        if (p100 == dataGapValue)
          p100 = Double.NaN;
        if (e06 == dataGapValue)
          e06 = Double.NaN;
        if (e2 == dataGapValue)
          e2 = Double.NaN;
        if (e4 == dataGapValue)
          e4 = Double.NaN;
        // Add data to vector.
        goesv.addElement(new GOES08part(d,p1,p5,p10,p30,p50,p100,e06,e2,e4,longitude));
      }
      // Trim the vector.
      goesv.trimToSize();
      // Close input stream.
      streamIn.close();
      // Store the data in the database
      Database db = Database.getInstance("GOES08partDB");

      ObjectTimeSeries ts = new ObjectTimeSeries(goesv);
      
      if(!db.isSet())
	System.out.println("MetaData for GOES08partDB is unavailable \nWill try to create it.");
      
      System.out.println("Got instance of DB will do store");
      db.store(ts);

    }
    catch (IOException e)
    {
      System.err.println("URL Connection error: " + e);
    }
    catch (NumberFormatException e)
    {
      System.err.println("Number format error: " + e);
    }
    catch (ParseException e)
    {
      System.err.println("Parse error: " + e);
    }
    catch (IllegalTimeSeriesException e) {
      System.err.println(e);
    }
    catch (Exception e) {
        System.err.println(e);
    }
  }
}
