package saaps.dbt;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.StreamTokenizer;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Calendar;
import java.util.TimeZone;

import saaps.util.ObjectTimeSeries;

/**
* Download the latest nowcasted and forecasted Kp values from
* the Lund site. The data are located at <br>
* http://sol.irfl.lu.se/KpForecast/latestNowcast.txt and <br>
* http://sol.irfl.lu.se/KpForecast/latestForecast.txt .
*
* @author Peter Wintoft
* @version 1.0
*/
public class LatestLundKp {

    public LatestLundKp() {
    }
    
    public void run() throws Exception {
        //Get the latest nowcasted Kp from Lund.
        Kp[] kp = new Kp[2];
	kp[0] = getLatestNowcast();
        //Get the latest forecasted Kp from Lund.
        kp[1] = getLatestForecast();
        //Store it in the database.
        this.store(kp);
    }
    
    private Kp getLatestNowcast() {
        Kp kp = this.getLatestPrediction("http://sol.irfl.lu.se/KpForecast/latestNowcast.txt");
        //Set the status to nowcast.
        kp.setStatus(3);
        return kp;
    }

    private Kp getLatestForecast() {
        Kp kp = this.getLatestPrediction("http://sol.irfl.lu.se/KpForecast/latestForecast.txt");
        //Set the status to forecast.
        kp.setStatus(4);
        return kp;
    }
    
    private Kp getLatestPrediction(String urlString) {
        Kp kp = null;
        //Read the data from the URL.
        try {
            URL url = new URL(urlString);
            InputStreamReader isr = new InputStreamReader(url.openStream());
            BufferedReader br = new BufferedReader(isr);
            StreamTokenizer st = new StreamTokenizer(br);
            Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
            cal.clear();
            //Set the year.
            st.nextToken();
            cal.set(Calendar.YEAR,(int) st.nval);
            //Set the month. Use -1 for the java month.
            st.nextToken();
            cal.set(Calendar.MONTH,(int) st.nval - 1);
            //Set the day.
            st.nextToken();
            cal.set(Calendar.DATE,(int) st.nval);
            //Set the hour.
            st.nextToken();
            cal.set(Calendar.HOUR_OF_DAY,(int) st.nval);
            //Read Kp.
            st.nextToken();
            double kpVal = st.nval;
            //Create Kp object using status 1.
            kp = new Kp(cal.getTime(),kpVal,1);
        }
        catch (MalformedURLException e) {
            System.err.println(e);
            System.exit(0);
        }
        catch (IOException e) {
            System.err.println(e);
            System.exit(0);
        }
        return kp;
    }
    
    private void store(Kp[] kp) throws Exception {
        //Connect to the database.
        Database db = Database.getInstance("KpDB");
        //Create the ObjectTimeSeries. When storing this into
        //the database the checks on correct sample time and
        //so on will be made.
        ObjectTimeSeries ots = new ObjectTimeSeries(kp);
        db.store(ots);
        //Close the database connection.
        try {
            db.close();
        }
        catch (IOException e) {
            System.err.println(e);
            System.exit(0);
        }
    }

    public static void main(String[] args) {
        //Create the application
        LatestLundKp kp = new LatestLundKp();
        //Run the application.
        try {
            kp.run();
        }
        catch (Exception e) {
            System.err.println(e);
        }
    }
}
