package saaps.network;

import Jama.Matrix;
import saaps.math.Math;

/**
The Neuron class is the smallest building block of
a network. A neuron is defined by the weights, the
bias and the activation function. The activation
functions currently supported are
<ul>
<li> PURELIN: y=x ,
<li> LOGSIG:  y=1/(1+exp(-x)) ,
<li> TANSIG:  y=2/(1+exp(-2x))-1 .
</ul>

@author Peter Wintoft
@version 2000-01-18
*/
public class Neuron
{
	// Weight matrix.
	private Matrix weight;
	// Bias.
	private double bias;
	// Activation function. Default to TANSIG.
	private int actFun = TANSIG;
	// The output matrix.
	private Matrix output;
	
	public final static int PURELIN = 0;
	public final static int LOGSIG = 1;
	public final static int TANSIG = 2;
	
	/**
	The neuron is created from the weights, the bias and
	the activation function.
	
	@param w weights
	@param b bias
	@param af activation function (PURELIN, LOGSIG or TANSIG)
	*/
	public Neuron(Matrix w, double b, int af)
	{
		weight = w;
		bias = b;
		actFun = af;
	}
	
	/**
	The neuron is created from the weights andthe bias using
	the default activation function (TANSIG).
	
	@param w weights
	@param b bias
	*/
	public Neuron(Matrix w, double b)
	{
		this(w,b,TANSIG);
	}
	
	/**
	Run the network on an input matrix. The number of rows should
	equal the number of inputs. Each column represents one example,
	and the output will have as many columns as the input. 
	
	@param input input matrix
	*/
	public Matrix sim(Matrix input)
	{
		Matrix b = new Matrix(1,input.getColumnDimension(),bias);
		switch (actFun)
		{
			case PURELIN:
				output =  weight.times(input).plus(b);
				break;
			case LOGSIG:
				output = Math.logsig(weight.times(input).plus(b));
				break;
		  default:
		  	output = Math.tansig(weight.times(input).plus(b));
		}
		return output;
	}
	
	/**
	Get the output.
	*/
	public Matrix getOutput()
	{
		return output;
	}
}