/*
 * BestPredModel.java
 *
 * Created on June 15, 2001, 3:48 PM
 */

package saaps.saam;

import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;

import saaps.math.Math;
import saaps.math.Prob;
import saaps.math.ProbCond;
import saaps.sapm.Anom002Model;
import saaps.util.EventList;
import saaps.util.Model;
import saaps.util.ModelList;
import saaps.util.TimeSeries;

/**
 *
 * @author  root
 * @version 
 */
public class BestPredModel extends java.lang.Object {
    
    //The models.
    private ModelList modelList;
    
    //Event list.
    private EventList eventList;

    //Non-Event list.
    private EventList nonEventList;
    
    //Observed and predicted outputs for each model.
    private double[][] observed;
    private double[][] predicted;
    
    public static final int LINCORR = 0;
    public static final int CONDPROB = 1;
    public static final int MUTUAL = 2;

    private static final String[] statString = {"Linear correlation","Conditional Probability"};
    
    //The type of statistics to use.
    private int usedStat = this.LINCORR;
    
    //The statistics.
    private double[] stat;
    
    public BestPredModel() {
        //Init the models.
        this.modelList = new ModelList();
    }
    
    //A calendar object for date calculations.
    private Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
    
    /** Creates new BestPredModel */
    public BestPredModel(EventList eventList, EventList nonEventList) {
        this.eventList = eventList;
        this.nonEventList = nonEventList;
        //Clean the event lists.
        this.eventList.useOnlyDates();
        this.nonEventList.useOnlyDates();
        //Init the models.
        this.modelList = new ModelList();
        //Init the output data.
        int n = this.eventList.getNumOfEvents() + this.nonEventList.getNumOfEvents();
        this.observed = new double[this.modelList.getNumOfModels()][n];
        this.predicted = new double[this.modelList.getNumOfModels()][n];
    }
    
    public void setEventList(EventList el) {
        this.eventList = el;
        this.eventList.useOnlyDates();
    }
    
    public void setNonEventList(EventList nel) {
        this.nonEventList = nel;
        this.nonEventList.useOnlyDates();
    }
        
//    private void cleanNonEventList() {
//        this.nonEventList.useOnlyDates();
//    }
        
    public EventList getEventList() {
        return this.eventList;
    }
    
    public EventList getNonEventList() {
        return this.nonEventList;
    }
    
    /**
     *This method must be called after <code>setEventList</code> or
     *<code>setNonEventList</code> has been called.
     */
    public void update() {
        //Init the output data.
        int n = this.eventList.getNumOfEvents() + this.nonEventList.getNumOfEvents();
        this.observed = new double[this.modelList.getNumOfModels()][n];
        this.predicted = new double[this.modelList.getNumOfModels()][n];
    }
    
    public void runModels() {
        //For each models calculate the predictions
        //for the events and non-events.
        for (int m=0;m<this.modelList.getNumOfModels();m++) {
            Model model = (Model) this.modelList.getModel(m);
            //The events.
            int numOfEvents = this.eventList.getNumOfEvents();
            for (int e=0;e<numOfEvents;e++) {
                //Get the prediction date. Assume that the date
                //in the event list marks the start time of the
                //day to be predicted, thus add one day to the 
                //prediction date.
                cal.setTime(this.eventList.getEvent(e));
                cal.add(Calendar.DATE,1);
                model.setPredictionTime(cal.getTime());
                //The model output is the probability in percent
                //that the prediction is an anomaly.
                model.run();
                TimeSeries ts = model.getOutput();
                this.observed[m][e] = 1.0;
                //The time series object containpredictions at
                //three hour sample interval. Use the last eight
                //values as the prediction for a day. If one or more
                //of the values are larger than 50% then it is an anomaly,
                //else it is a no-anomaly.
                this.predicted[m][e] = 0.0;
                for (int i=ts.getLength()-8;i<ts.getLength();i++)
                    if (ts.getData(i,0) > 50.0) this.predicted[m][e] = 1.0;
            }
            //The non-events.
            for (int ne=0;ne<this.nonEventList.getNumOfEvents();ne++) {
                //Get the prediction date. Assume that the date
                //in the event list marks the start time of the
                //day to be predicted, thus add one day to the 
                //prediction date.
                cal.setTime(this.nonEventList.getEvent(ne));
                cal.add(Calendar.DATE,1);
                model.setPredictionTime(cal.getTime());
                //The model output is the probability in percent
                //that the prediction is an anomaly.
                model.run();
                TimeSeries ts = model.getOutput();
                this.observed[m][ne + numOfEvents] = 0.0;
                //The time series object containpredictions at
                //three hour sample interval. Use the last eight
                //values as the prediction for a day. If one or more
                //of the values are larger than 50% then it is an anomaly,
                //else it is a no-anomaly.
                this.predicted[m][ne + numOfEvents] = 0.0;
                for (int i=ts.getLength()-8;i<ts.getLength();i++)
                    if (ts.getData(i,0) > 50.0) this.predicted[m][ne + numOfEvents] = 1.0;
            }
        }
    }
    
    public void setStatistics(int s) {
        this.usedStat = s;
    }
    
    public int getStatistics() {
        return this.usedStat;
    }
    
    public double getStatValue(int i) {
        return this.stat[i];
    }
    
    /**
     *Compute the statistics for each model. A double array is returned
     *with the number of elements equal to the number of models.
     *
     *@return A double array with the statistics.
     */
    public void computeStatistics() {
        if (this.usedStat == this.CONDPROB) {
                this.computeCondProb();
        }
        if (this.usedStat == this.LINCORR) {
                this.computeLinCorr();
        }
    }
    
    private void computeLinCorr() {
        this.stat = new double[this.modelList.getNumOfModels()];
        for (int m=0;m<this.modelList.getNumOfModels();m++) {
            this.stat[m] = Math.linCorr(this.observed[m],this.predicted[m]);
        }
    }
    
    private void computeCondProb() {
        this.stat = new double[this.modelList.getNumOfModels()];
        int[][] x = new int[this.observed.length][this.observed[0].length];
        int[][] y = new int[this.predicted.length][this.predicted[0].length];
        for (int m=0;m<x.length;m++) {
            for (int e=0;e<x[0].length;e++) {
                x[m][e] = (int) java.lang.Math.round(this.observed[m][e]);
                y[m][e] = (int) java.lang.Math.round(this.predicted[m][e]);
            }
        }
        Prob prob = new Prob();
        ProbCond cond = new ProbCond();
        for (int m=0;m<this.modelList.getNumOfModels();m++) {
            prob.calcProb(x[m],y[m]);
            cond.calcProbCond(prob);
//System.out.println(m);
//            this.stat[m] = 2.0;
            this.stat[m] = (cond.xy[0][0] + cond.xy[1][1]) / 2.0;
        }
    }
    
    public void sortModels() {
        this.modelList.sort(this.stat);
    }
    
    public ModelList getModelList() {
        return this.modelList;
    }
    
    public int getNumOfModels() {
        return this.modelList.getNumOfModels();
    }
    
    public String getModelName(int i) {
        return this.modelList.getModel(i).getName();
    }
    
    public String toString(int m) {
        java.text.DecimalFormat df = new java.text.DecimalFormat("0.00");
        String div = "\n---------------------------------";
        String s = "Model :" + this.modelList.getModel(m).getName();
        s += div;
        int e = this.eventList.getNumOfEvents();
        int ne = this.nonEventList.getNumOfEvents();
        int tot = e + ne;
        s += "\nTotal number of events: " + tot;
        s += "\nNumber of 'no anomaly': " + ne;
        s += "\nNumber of 'anomaly'   : " + e;
        s += div;
        s += "\nResult:";
        s += "\n" + this.statString[this.usedStat] + "=";
        if (Double.isNaN(this.stat[m]))
            s += "NaN";
        else
            s += df.format(this.stat[m]);
        s += "\n";
        //Result text.
        String weak = "\nThe model predicts the submitted data poorly." +
                      "\nThe anomalies may not be related to the type" +
                      "\nof anomalies predicted by the model. See the" +
                      "\nmodel information below.";
        String medium = "\nThere is some predictability of the submitted data." +
                        "\nA subset of anomalies may be related to the type" +
                        "\nof anomalies predicted by the model, while another" +
                        "\nsubset is not. See the model information below";
        String strong = "\nThe model predicts the submitted data well." +
                        "\nThe anomalies may be related to the type" +
                        "\nof anomalies predicted by the model. See the" +
                        "\nmodel information below.";
        //Linear correlation.
        if (this.usedStat==this.LINCORR) {
            //Weak correlation.
            if (this.stat[m]<0.2) {
                s += weak;
            }
            //Medium correlation.
            if (this.stat[m]>=0.2 & this.stat[m]<0.7) {
                s += medium;
            }
            //Strong correlation.
            if (this.stat[m]>=0.7) {
                s += strong;
            }
        }
        //Cond. prob.
        if (this.usedStat==this.CONDPROB) {
            //Weak correlation.
            if (this.stat[m]<0.6) {
                s += weak;
            }
            //Medium correlation.
            if (this.stat[m]>=0.6 & this.stat[m]<0.85) {
                s += medium;
            }
            //Strong correlation.
            if (this.stat[m]>=0.85) {
                s += strong;
            }
        }
        //Model information.
        s += div;
        s += "\n" + this.modelList.getModel(m).getInfo();
        //Help information.
        s += div;
        s += "\nFor more information see the help pages.";
        return  s;
    }
    
    public String toString() {
        StringBuffer buf = new StringBuffer();
        for (int m=0;m<this.stat.length;m++) {
            buf.append("Model " + m + " : " + this.stat[m] + "\n");
        }
        return buf.toString();
    }

    /**
    * @param args the command line arguments
    */
    public static void main (String args[]) {
        //Get the event list.
        System.out.println("Reading the event list ...");
        EventList events = new EventList();
        events.parseEventFile(args[0]);
        System.out.println(events);
        //Get the non-event list
        System.out.println("Reading the non-event list ...");
        EventList nonEvents = new EventList();
        nonEvents.parseEventFile(args[1]);
        System.out.println(nonEvents);
        //Create the BestPredModel object.
        System.out.println("Creating the BestPredModel object from the event lists ...");
        BestPredModel bpm = new BestPredModel(events,nonEvents);
        //Run the models.
        System.out.println("Running the models ...");
        bpm.runModels();
        //Compute the statistics.
        bpm.computeStatistics();
        //Display the result.
        System.out.println(bpm);
    }

}
