package saaps.util;

/**
The ObjectTimeSeries class is a storage bin for DatabaseObjects.
In here are also methods to check the data according rules 
derived from the database where they are stored (MetaData).

@author Leif Kalla
@version 2000-03-30
*/
import saaps.dbt.*;

import java.util.*;
import java.text.SimpleDateFormat;
import java.text.NumberFormat;
import java.io.Serializable;
import java.io.IOException;

import Jama.Matrix;

public class ObjectTimeSeries implements Serializable {
    // dates contain the dates for each observation
    private DatabaseObject[] items;
    // the number of data.
    private int numOfData;
    // the number of fields.
    private int numOfFields;
    // the number of data fields.
    private int numOfDataFields;
    // Time zone = UT.
    private TimeZone tz = TimeZone.getTimeZone("GMT");
    private Calendar cal = Calendar.getInstance(tz);

    /**
   * Create an ObjectTimeSeries from an array of DatabaseObjects.
   */
    public ObjectTimeSeries(DatabaseObject[] objArr) {
	numOfData = objArr.length;
	numOfFields = ((DatabaseObject) objArr[0]).getNumberOfDataIndexes() +1;
	numOfDataFields = numOfFields - 1;

	items = objArr;
    }

    /**
  Create a ObjectTimeSeries object from a Vector of DatabaseObject:s

  @param dbov The DatabaseObject vector
  */
    public ObjectTimeSeries(Vector dbov) throws IllegalTimeSeriesException {

	// Get the number of data and the number of data fields.
	//Assume that the vector contains same objects all the way
	//which should be true !!!
	numOfData = dbov.size();
	if(numOfData == 0) throw new IllegalTimeSeriesException("Size is 0");
	numOfFields = ((DatabaseObject) dbov.elementAt(0)).getNumberOfDataIndexes() +1;
	numOfDataFields = numOfFields - 1;

	//Create a matrix and fill it, fill the vector
	items = new DatabaseObject[numOfData];
	dbov.copyInto(items);
    }

    /*
  A method to set a Calendar object from a Date object
  and setting the hour, minute, and second fields all to
  zero.

  @param d the date
  */
    private Calendar setCalDay(Date d) {
	Calendar cal = Calendar.getInstance(tz);
	cal.setTime(d);
	cal.set(cal.HOUR,0);
	cal.set(cal.MINUTE,0);
	cal.set(cal.SECOND,0);
	cal.set(cal.MILLISECOND,0);
	return cal;
    }

    /**
  Get the length of the ObjectTimeSeries object.
  */
    public int getLength() {
	return numOfData;
    }

    /**
  Get one Date object.
  */
    public Date getDate(int i) {
	return items[i].getDate();
    }

    /**
   * Get one data Object
   */
    public DatabaseObject getDataObject(int i) {
	return items[i];
    }

    /**
   * Get first object.
   * @return the first object
   */
    public DatabaseObject getFirst() {
	return items[0];
    }

    /**
     * Get last object
     * @return the last object
     */
    public DatabaseObject getLast() {
	return items[numOfData-1];
    }

    /**
  Get the number of data fields.
  */
    public int getNumOfDataFields() {
	return numOfDataFields;
    }

    /**
   * Run this to check if the time series is valid according to 
   * the database it's supposed to be in.
   * If it's not we return false ...
   */
    public boolean checkDbValid(MetaData meta) 
	throws IllegalTimeSeriesException ,
	TimeSeriesSequenceException ,
	TimeSeriesContinuityException {

	//Check that first time is not before start of data in DB
	if(items[0].getDate().before(meta.getStartCalendar().getTime()))
	    throw new IllegalTimeSeriesException("First time is before start of Database");

	//Check that first time is not after last time of data in DB
        //+meta.getTimeStep()
	if(items[0].getDate().after(new Date(meta.getEndCalendar().getTime().getTime()))) 
	    throw new TimeSeriesEODException("First time is after end of Database");


	//From here first data value should be OK it's inside 
	if(meta.getCont()) { //If should be continous check it
	    for(int i=1;i<numOfData;i++) {
		//System.out.println("debug:" + meta.getTimeStep() + " " + (dates[i].getTime()-dates[i-1].getTime()));
		if((items[i].getDate().getTime()-items[i-1].getDate().getTime()) != meta.getTimeStep())
		    throw new TimeSeriesContinuityException("TimeSeries not continous");
	    }
    
	}
	else { 
	    //If not continous , at least it should go into the future ...
	    for(int i=1;i<numOfData;i++) {
		if((items[i].getDate().getTime()-items[i-1].getDate().getTime()) < 0)
		    throw new TimeSeriesSequenceException("Time series not a correct sequence");
	    }
	}
	return true;
    }

    /**
     * Print info. about data to standard output
     */
    public void print() {

	SimpleDateFormat formatter = (SimpleDateFormat)
	    SimpleDateFormat.getDateTimeInstance(0,0,new Locale("sv","SE"));
 
	formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
	formatter.applyPattern("yyyy.MM.dd HH:mm:ss");

	NumberFormat nf = NumberFormat.getInstance();
	nf.setMaximumFractionDigits(3);
	nf.setMinimumFractionDigits(3);

	for(int i=0;i<items.length;i++) {
	    System.out.print(items[i].getClass().getName() + "|" +
			     formatter.format(items[i].getDate()));
	    for(int j=0;j<numOfDataFields;j++) {
		System.out.print("|" + nf.format(items[i].getData(j)));
	    }
	    System.out.println("");
	}
    }

    /**
     * Averages all fields in the data and returns a new 
     * <code>DatabaseObject</code>. The time is set to the middle of
     * the data.
     * @return the averaged object
     */
    public DatabaseObject average() {

	double[] data = new double[numOfDataFields];
	int cnt[] = new int[numOfDataFields]; //Count the number of data points

	/* Get the ave time */
	Date date = null;

	/* if only one object return it */
	if(items.length == 1)
	    return items[0];

	/* if odd number of objects pick the one in the middle */
	if(items.length%2 != 0) {
	    date = ((DatabaseObject) items[(int) (items.length-1)/2]).getDate();
	}
	else {
	    int pos = ((int) items.length/2) - 1 ;    //position at date before middle
	    Date date1 = ((DatabaseObject) items[pos]).getDate();
	    pos++;     //position after middle
	    Date date2 = ((DatabaseObject) items[pos]).getDate();
	    date = new Date(date1.getTime() + (date2.getTime()-date1.getTime())/2);
	}

	for(int i=0;i<items.length;i++) {
	    for(int j=0;j<numOfDataFields;j++) {
		double tmp = ((DatabaseObject)items[i]).getData(j);
		/* Skip NaN:s */
		if(!Double.isNaN(tmp)) {
		    data[j]+= tmp;
		    cnt[j]++;
		}
	    }
	}

	/* Average */
	for(int j=0;j<numOfDataFields;j++) {
	    data[j]/= cnt[j];
	}

	String name = items[0].getClass().getName();

	return (saaps.dbt.DatabaseObject.
		getInstance(name.substring(name.lastIndexOf('.')+1),date,data));
    }

    /**
     * Averages all fields in the data for the selected time interval and 
     * returns a new <code>DatabaseObject</code>.
     * The time is set to the sample time.
     * @param sampleTime time of the resulting object.
     * @param wa time offset 1 (negative value means before sampleTime)
     * @param wb time offset 2 (positive value means after sampleTime)
     * @return the averaged object
     */
    public DatabaseObject average(Date sampleTime,long wa,long wb) {

	double[] data = new double[numOfDataFields];
	int cnt[] = new int[numOfDataFields]; //Count the number of data points

	/* set time of resulting object */
	Date date = sampleTime;

	/* if only one object return it */
	if(items.length == 1)
	    return items[0];
	Date startTime = new Date(sampleTime.getTime() + wa);
	Date endTime = new Date(sampleTime.getTime() + wb);

	if(startTime.after(endTime)) {
	    System.err.println("starttime is after endtime.");
	    return null;
	}
	
	for(int i=0;i<items.length;i++) {
	    /* if inside time interval use it else do nothing */
	    if((items[i].getDate().compareTo(startTime) >= 0) && (items[i].getDate().compareTo(endTime) <= 0)) {
		for(int j=0;j<numOfDataFields;j++) {
		    double tmp = ((DatabaseObject)items[i]).getData(j);
		    /* Skip NaN:s */
		    if(!Double.isNaN(tmp)) {
			data[j]+= tmp;
			cnt[j]++;
		    }
		}
	    }
            else if(items[i].getDate().after(endTime))
                break;
	}

	/* Average */
	for(int j=0;j<numOfDataFields;j++) {
	    data[j]/= cnt[j];
	}

	String name = items[0].getClass().getName();

	return (saaps.dbt.DatabaseObject.
		getInstance(name.substring(name.lastIndexOf('.')+1),date,data));
    }

    /**
   * Run this to create an ObjectTimeSeries with empty data values
   * for the dates in between lastInDb and first in ObjectTimeSeries.
   */
    public ObjectTimeSeries fitToDatabase(MetaData meta,Date lastInDb) 
	throws IllegalTimeSeriesException {
	int i=0; //our loop counter 

	if(!meta.getCont())
	    throw new IllegalTimeSeriesException("For non continous databases we don't need this.");

	//Calculate how many items we have to insert.
        //System.out.println("lastInDb:"+lastInDb + " timeStep:"+meta.getTimeStep()+" ots first:"+items[0].getDate());
	int noEmpty = ((int) ((items[0].getDate().getTime() - lastInDb.getTime())/((long) meta.getTimeStep()))) - 1;

	//Play it safe ??? 43200 is one sample every minute for 30 days
	if(noEmpty > 43200)
	    throw new IllegalTimeSeriesException("Too many DatabaseObjects needed:" + noEmpty);
        if(noEmpty <= 0)
            throw new IllegalTimeSeriesException("Too few DatabaseObjects needed:" + noEmpty);
        
	DatabaseObject[] nDboArr= new DatabaseObject[noEmpty+getLength()];
	String objectName = meta.getName().substring(0, meta.getName().length()-2);

	//Get a calendar so that we can create the needed Date objects
	Calendar tmpCal=Calendar.getInstance(TimeZone.getTimeZone("GMT"),new Locale("sv","SE"));
	tmpCal.setTime(lastInDb); //Increase this with the timeStep in MetaData

	//OK start filling the array with the correct objects
	//The empty ones
	for(i=0;i<noEmpty;i++) {
	    tmpCal.add(Calendar.MILLISECOND,(int) meta.getTimeStep()); //millisec
	    nDboArr[i] = DatabaseObject.getInstance(objectName,tmpCal.getTime());
	}
	System.out.println("fit:" + i + "noEmpty " + noEmpty + "Length " + getLength());
	//The rest which is in the current ObjectTimeSeries object
	for(;i<(noEmpty+getLength());i++) {
	    nDboArr[i] = getDataObject(i-noEmpty);
	}

	//Ok get the new object and return it
	return new ObjectTimeSeries(nDboArr);
    }

    /**
   * Run this to create an ObjectTimeSeries with empty data values
   * for the data gaps.
   */
    public static Vector fillDataGaps(Vector dbov,MetaData meta) 
	throws IllegalTimeSeriesException {

	SimpleDateFormat formatter = (SimpleDateFormat) SimpleDateFormat.getDateTimeInstance(0,0,new Locale("sv","SE"));

	formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
	formatter.applyPattern("yyyy MM dd HH mm ss");

	int i=0,j=0; //our loop counters

	if(!meta.getCont())
	    throw new IllegalTimeSeriesException("For non continous databases we don't need this.");

	String objectName = meta.getName().substring(0, meta.getName().length()-2);
	//Get a calendar so that we can create the needed Date objects
	Calendar tmpCal=Calendar.getInstance(TimeZone.getTimeZone("GMT"),new Locale("sv","SE"));
	//Search for data gaps
	for(i=1;i<dbov.size();i++) {
	    long timeDiff = ((DatabaseObject) dbov.elementAt(i)).getDate().getTime()- ((DatabaseObject) dbov.elementAt(i-1)).getDate().getTime(); //ms
	    if( timeDiff != meta.getTimeStep()) {
		if( timeDiff < 0 ) { // Time jumps backward, we don't correct that type of errors
		    System.out.println("fillDataGaps t2:" + formatter.format(((DatabaseObject) dbov.elementAt(i)).getDate()) + " t1:" + formatter.format(((DatabaseObject) dbov.elementAt(i-1)).getDate()) + " timeDiff:" + timeDiff);
		    throw new IllegalTimeSeriesException("Time jumps backward !!");
		}
		int addThisMany = (int) (timeDiff/meta.getTimeStep());
		System.out.println("will add data from t1:" + formatter.format(((DatabaseObject) dbov.elementAt(i-1)).getDate()) + " to t2:" + formatter.format(((DatabaseObject) dbov.elementAt(i)).getDate()) + " num items:" + addThisMany);
		if(addThisMany > 43200)
		    throw new IllegalTimeSeriesException("Too many DatabaseObjects needed:" + addThisMany);

		tmpCal.setTime(((DatabaseObject) dbov.elementAt(i)).getDate()); //Increase this with the timeStep in MetaData

		//OK start filling the array with the correct objects
		//The empty ones
		for(j=0;j<addThisMany;j++) {
		    tmpCal.add(Calendar.MILLISECOND,(int) meta.getTimeStep()); //millisec
		    //Insert an object at this point
		    dbov.add(i,DatabaseObject.getInstance(objectName,tmpCal.getTime()));
		    i++; //Increase the dbov counter
		}
		i--; //go back one for the check
	    }
	}
	//Return the modified object reference, actually you should just use the 
	//reference you passed here (the object itself is modified).
	return (dbov);
    }

    /**
   * Resamples all fields in the data.
   *
   * XXX The list of objects has to be scanned through for every new
   * sample.
   *
   * @param dt0 offset time from the first time in the series [ms]
   * @param dt new sample interval (timestep [ms])
   * @param wa time offset where resampling begins [ms]
   * @param wb time offset where resampling ends [ms]
   * @return -1 if failed 0 when normal operation.
   */
    public byte resample(long dt0,long dt,long wa,long wb) {

	double[] data = new double[numOfDataFields];
	int cnt[] = new int[numOfDataFields]; //Count the number of data points

	/* Get the ave time */
	Date date = null;

	long timeSpan = getLast().getDate().getTime() - getFirst().getDate().getTime();
	System.out.println("TimeSpan:" + timeSpan);
	/* Create a list that can hold the resampled data objects */
	int objCnt = (int)( timeSpan/dt);
	if(objCnt < 2) {
	    System.err.println("List is too short:" + objCnt);
	    return -1;
	}
	System.out.println("num new objs:" + objCnt);

	DatabaseObject[] newList = new DatabaseObject[objCnt];
	//First sample will be here
	Calendar mtime = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
	mtime.setTime(new Date(items[0].getDate().getTime()+dt0));

	for(int i=0;i<objCnt;i++) {
	    newList[i] = average(mtime.getTime(),wa,wb);
	    //step time forward with the new sample interval
	    mtime.add(Calendar.SECOND,(int) dt/1000);
	    mtime.add(Calendar.MILLISECOND,(int) (dt - ((int)dt/1000)*1000));
	}

	items = newList;
	numOfData = items.length;
	//numOfFields = should be the same because it's the same data type !!
	//numOfDataFields = should be the same because it's the same data type !!

	return 0;
    }

    static public void main(String[] args) {

	try {
	    NumberFormat nf = NumberFormat.getInstance();
	    nf.setMaximumFractionDigits(3);
	    nf.setMinimumFractionDigits(3);

	    Calendar sd = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
	    Calendar ed = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
	    sd.set(2000,6-1,1,0,0,0);
	    sd.clear(sd.MILLISECOND);
	    ed.set(2000,6-1,2,0,0,0);
	    ed.clear(ed.MILLISECOND);
	    long start = -10*60*1000;
	    long end = 10*60*1000;

	    Database db = Database.getInstance(args[0],true);

	    ObjectTimeSeries ots = db.retrieveOTS( sd.getTime(), ed.getTime(),false);
	    ots.print();
	    ots.resample(0,60*60*1000,-30*60*1000,30*60*1000);
	    ots.print();
	    
	    TimeSeries ts = new TimeSeries(ots,db.meta.getFieldNames(),db.meta.getFieldUnits());
	    System.out.println("TimeSeries: " + ts.getShortDescription() + " long descr:" + ts.getLongDescription());
	    
	    String[] fn = ts.getFieldNames();
	    for(int i=0;i<fn.length;i++) {
		System.out.print(fn[i]);
		if(i<(fn.length-1))
		    System.out.print(" : ");
	    }
	    System.out.println("");
 
	    for(int i = 0;i<ts.getLength();i++) {
		System.out.print("d: " + ts.getDate(i));
 
		for(int j=0;j<ts.getNumOfDataFields();j++) {
		    System.out.print(" " + nf.format(ts.getData(i,j)));
		}
		System.out.println("");
	    }

	}
	catch (Exception e) {
	    System.err.println(e);
	    e.printStackTrace();
	}
    }
}

