#!/usr/bin/perl -w
###
# $Id: sedat_build_query.pl,v 1.6 2004/12/15 23:14:00 stamper Exp $
#
#  Purpose
#
#   To build a query then control its execution.
#
#  Arguments/Operator input
#
#   Argument       Description
#   --------------------------
#   $user          The user running the query
#   $query         The query to be run
#
#  External routines
#
#   Sedat.pm       Several utility routines in module Sedat.
#   send_idl       A C program to send commands to the IDL server
#
#  Description
#
#  Author          Richard Stamper
#
# $Log: sedat_build_query.pl,v $
# Revision 1.6  2004/12/15 23:14:00  stamper
# Append system error message to the abort message when failing after
# trying to modify a file.
#
# Revision 1.5  2002/12/05 15:21:57  stamper
# Display the query input arguments names and values at the top of the
# query log file.
#
# Revision 1.4  2001/12/06 15:33:23  stamper
# Do not truncate script lines when a semi-colon is found, since that might
# be inside a string.
#
# Revision 1.2  2000/10/16 23:49:53  stamper
# Match 'D' and 'V' characters indicating input/output types only at the
# beginning of a string.
#
# Revision 1.1  2000/08/03 14:37:14  stamper
# First checkpoint.
#
#
###

use Sedat;
use English;
use Getopt::Std;

$mapFileIndex = 'fh00';

#
# Process inputs - user, query
#
getopts('u:,q:');

#
# check user given
#
if (!defined($opt_u))
{
    abortQuery("No user");
}
#
# check query name given
#
if (!defined($opt_q))
{
    abortQuery("No query");
}

#
# check user exists in sedat structure
#
$user = $opt_u;
$userDir = objectName($user,$user,'U',$user);
if (! -d $userDir)
{
    abortQuery("User $user does not exist in SEDAT structure");
}
$IDL_dir = joinName($userDir,'idl');
$systemDir = joinName(objectName('SYSTEM','SYSTEM','U','SYSTEM'),'idl');

#
# check query exists in user's sedat structure
#
$query = $opt_q;
$queryDir = objectName($user,$user,'Q',$query);
if (! -d $queryDir)
{
    abortQuery("Query $queryDir does not exist in ${user}'s SEDAT structure");
}

#
# check status file of query
#
$statusFile = joinName($queryDir,'status');
if (! -e $statusFile)
{
    abortQuery("No status file for query $query");
}
if (-s $statusFile == 0)
{
    abortQuery("No information in status file for query $query");
}

#
# process mapping file of query
#
$mappingFile = joinName($queryDir,'mapping');
if (checkMapFile($mappingFile) eq 'library')
{
    abortQuery("Can't have library mapping for query");
}

#
# check status
#
if (checkStatus($statusFile) ne 'Y')
{
    abortQuery("Query $query is not executable");
}

#
# generate query instance name
#
$queryInstance = $query . "_" . timeStamp(time());
# add query instance name to query structure
$queryStruct->{QUERYNAME} = $queryInstance;

# Create the query instance directory
$queryInstanceDir = joinName($queryDir,'instances',$queryInstance);
mkdir $queryInstanceDir,0775 or
    abortQuery("Failed to create query instance directory $queryInstanceDir:\n$!");

# Set up commands file details then create the commands file
$queryCmds = joinName($queryInstanceDir,'commands');
open QUERYCMDS,">$queryCmds" or
    abortQuery("Failed to create query commands file $queryCmds\n$!");

# add users idl directory to IDL path
IDLexecute("!path='${systemDir}:${IDL_dir}:'+!path");
# set up the IDL library
IDLexecute('s = call_external("libidl_stpdf.so","idl_chp_open")');
# compile ODI_tools
IDLexecute(".COMPILE odi_tools");

#
# process query mapping file
#

# Initialise the record of input arguments
@query_args = ();
$arg_index  = 0;
#
# open query mapping file
#
$mapFileIndex++;
open($mapFileIndex, $mappingFile);
# process each line of query mapping file
while (<$mapFileIndex>)
{
    # split mapping file line
    ($toolInfo, $queryOutputs, $datasetList) =
        split(/;\s*/, $_, 3);
    # split tool information
    ($toolNum, $toolOwner, $toolName) =
        split(/!/, $toolInfo);
    # dummy use of toolNum
    if ($toolNum > 1)
    {
        exit(0);
    }
    # add tool name and owner to query structure
    $queryStruct->{TOOLNAME}  = $toolName;
    $queryStruct->{TOOLOWNER} = $toolOwner;
    $queryStruct->{QUERYDIR}  = $queryDir;
    # get tool
    $toolFile = getToolName($toolOwner, $toolName);
    $toolStatus = joinName($toolFile,'status');
    if (checkStatus($toolStatus) ne 'Y')
    {
        abortQuery("Tool $toolFile is not executable");
    }

    # process query output files
    ($outputTypes, $outputComposite) =
        split(/:/,$queryOutputs);
    $outputTypes = '' if ! defined($outputTypes);
    $outputNum = length($outputTypes);
    for ($ofile=1; $ofile<=$outputNum; $ofile++)
    {
        $outputFiles[$ofile-1] = "${queryInstance}_${ofile}";
        # check if temporary file
        if (substr($outputComposite,$ofile-1,1) ne 'Y')
        {
            $outputFiles[$ofile-1] .= ".tmp";
        }
    }
    $queryStruct->{OUTPUTFILES} = [@outputFiles];
    # process query input files
    # find datasets
    $datasetList = '' if !defined($datasetList);
    @datasetlist = split(/;\s*/, $datasetList);
    chomp @datasetlist;

    $ifile = 0;
    $ivec = 0;
    while (@datasetlist)
    {
        ($datatype, $dataset) = split(/:/, shift(@datasetlist),2);

        foreach ($datatype)
        {
	    $arg_name = substr($datatype,2);
	    $arg_name = "Argument $arg_index" if ($arg_name eq '');

            /^V/ && do # Vector
            {
                # add vector to IDL structure
                $inputVectors[$ivec] = substr($dataset,1,length($dataset)-2);
		push @query_args,
		     sprintf('%s = %s',$arg_name,$inputVectors[$ivec]);

                $ivec++;

                last;
            };
            /^D/ && do # Datafile
            {
                ($dataowner, $dataname) = split(/!/, $dataset);
                if ($dataowner eq "SYSTEM")
                {
                    $datafile = $dataset;
                    $datadir  = 'LRD';
                    $realDataFile = objectRealName($dataowner,'D',$dataname);
                }
                else
                {
                    $datafile = 'data';
                    $datadir  = objectName($user,$dataowner,'D',$dataname);
                    $realDataFile = $datadir;
                }
                if (! -e $realDataFile)
                {
                    abortQuery("Dataset $dataowner\!$dataname not available");
                }

                # add data input file to IDL structure
                $inputDataFiles[$ifile] = $datafile;
                $inputDataDirs[$ifile]  = $datadir;
                $ifile++;
		push @query_args, sprintf('%s = %s',$arg_name,$dataset);

                last;
            };
	    $arg_index++;
        }
    }
    $queryStruct->{INPUTVECTORS} = [@inputVectors];
    $queryStruct->{INPUTFILES}   = [@inputDataFiles];
    $queryStruct->{INPUTDIRS}    = [@inputDataDirs];

    $toolMapFile = joinName($toolFile,'mapping');
    foreach (checkMapFile($toolMapFile))
    {
        /tool/ && do
        {
            processToolMapping($queryInstance,$toolMapFile,$mapFileIndex,
                               $queryStruct);
            last;
        };
        /library/ && do
        {
            # prepare IDL library and copy to users IDL directory
            $toolScriptFile = joinName($toolfile,'script');
            $toolIDLFile = "${queryInstance}_${toolname}_${toolnum}";
            $toolStruct->{SCRIPT} = $toolScriptFile;
            $toolStruct->{IDL} = lc($toolIDLFile);

            # need to prepare IDL, without adding PRO etc. to front
            prepareIDL($toolStruct,'library');
            # compile library code
            compileIDL($toolStruct);

            last;
        };
    }

}
# close query mapping file
close $mapFileIndex;

# Close the commands file
close QUERYCMDS;

# Create a status file
$queryStat = joinName($queryInstanceDir,'stat');
open QUERYSTAT,">$queryStat" or
    abortQuery("Failed to create query status file $queryStat\n$!");
print QUERYSTAT "Queued";
close QUERYSTAT;

print STDOUT $queryInstance;
exit(0);

sub timeStamp
{
    my ($time) = @_;
    my ($null);

    # get time details
    ($sec,$min,$hr,$mday,$mon,$yr,$null) = localtime($time);
    # add correction for returned number of years since 1900
    $yr = $yr + 1900;
    # add correction for returned month in range 0-11
    $mon = $mon + 1;
    $timestamp = sprintf("%4d%02d%02d%02d%02d%02d",
                         $yr, $mon, $mday, $hr, $min, $sec);
    return $timestamp;
}

sub checkStatus
{
    my ($statusfile) = @_;

#
# open status file
#
    open STATUSFILE, $statusfile;

# clear status variables
    $status_executable = 'N';
    $status_keywords = '';
    $status_released = '';
    $status_input = '';
    $status_output = '';

# process status file
    while(<STATUSFILE>)
    {
        ($flag, $flagstat) = split(/:\s*/, $_);
        chomp($flagstat);
        foreach ($flag)
        {
            /script-valid/ && do
            {
                $status_executable = $flagstat;
                last;
            };
            /keywords/ && do
            {
                $status_keywords = $flagstat;
                last;
            };
            /released/ && do
            {
                $status_released = $flagstat;
                last;
            };
            /'input'/ && do
            {
                $status_input = $flagstat;
                last;
            };
            /'output'/    && do
            {
                $status_output = $flagstat;
                last;
            };
        }
    }

# close status file
    close STATUSFILE;

    return $status_executable;
}

sub checkMapFile
{
    my ($mapFile) = @_;
    my ($mapType);

    if (! -e $mapFile)
    {
        abortQuery("No mapping file for query $query");
    }
    if (-s $mapFile == 0)
    {
        $mapType = 'library';
    }
    else
    {
        $mapType = 'tool';
    }

    return $mapType;
}

sub processToolMapping
{
    my ($query_instance, $input_mappingfile, $mapFileIndex,
        $compositeStruct) = @_;
    my ($statusfile, $tool_mappingfile, $toolType, $toolStruct);

    # transfer relevant items to tool structure from composite structure
    $toolStruct->{QUERYNAME} = $compositeStruct->{QUERYNAME};
    $toolStruct->{CTOOLNAME} = $compositeStruct->{TOOLNAME};
    $toolStruct->{QUERYDIR}  = $compositeStruct->{QUERYDIR};

    $mapFileIndex++;
    open($mapFileIndex, $input_mappingfile);
    while (<$mapFileIndex>)
    {
# find tool owner!name
        ($toolinfo, $outputs, $datasetlist) =
            split(/;\s*/, $_, 3);
        ($toolnum, $toolowner, $toolname) =
            split(/!/, $toolinfo);
        # add tool name to tool structure
        if ($toolowner eq '[this]')
        {
            $toolname = $compositeStruct->{TOOLNAME};
            $toolStruct->{TOOLNAME}  = $compositeStruct->{TOOLNAME};
            $toolStruct->{TOOLOWNER} = $compositeStruct->{TOOLOWNER};
        }
        else
        {
            $toolStruct->{TOOLNAME}  = $toolname;
            $toolStruct->{TOOLOWNER} = $toolowner;
        }

        # add input files to tool structure
        $toolStruct->{INPUTFILES} = $compositeStruct->{INPUTFILES};
        $toolStruct->{INPUTDIRS}  = $compositeStruct->{INPUTDIRS};
        # add input vectors to tool structure
        $toolStruct->{INPUTVECTORS} = $compositeStruct->{INPUTVECTORS};

        $toolfile = getToolName($toolStruct->{TOOLOWNER}, $toolname);
        $tool_mappingfile = joinName($toolfile,'mapping');
        # check whether tools mapping file exists and hence find type
        # tool or library
        $toolType = checkMapFile($tool_mappingfile);
        if ($toolType eq 'library') {
            # prepare IDL library and copy to users IDL directory
            $toolScriptFile = joinName($toolfile,'script');
            $toolIDLFile = "${query_instance}_${toolname}_${toolnum}";
            $toolStruct->{SCRIPT} = $toolScriptFile;
            $toolStruct->{IDL} = lc($toolIDLFile);

            # need to prepare IDL, without adding PRO etc. to front
            prepareIDL($toolStruct,'library');
            # compile library code
            compileIDL($toolStruct);
        } elsif ($toolType eq 'tool') {
            # check outputs
            ($output_types, $output_composite) =
                split(/:/, $outputs);
            # check number of outputs
            $output_types = '' if !defined($output_types);
            $output_num = length($output_types);
            # create data file name for each output file
            for ($ofile=1; $ofile<=$output_num; $ofile++) {
                if (substr($output_composite,$ofile-1,1) eq 'Y') {
                    $outputfilename[$ofile-1] =
                        shift @ { $compositeStruct->{OUTPUTFILES} };
                } else {
                    $outputfilename[$ofile-1] =
                        sprintf("$compositeStruct->{QUERYNAME}_${toolname}_${toolnum}_${ofile}");
                }
            }
            # add output data files to IDL structure
            $toolStruct->{OUTPUTFILES} = [@outputfilename];

            # build dataset structures/hashes
            $datasetlist = '' if !defined($datasetlist);
            processDatasetList($datasetlist, $toolStruct);

            if ($toolowner eq '[this]') {
                $toolfile = getToolName($toolStruct->{TOOLOWNER}, $toolname);
                # copy tool to users IDL directory
                $toolScriptFile = joinName($toolfile,'script');
                $toolIDLFile = "${query_instance}_${toolname}_${toolnum}";
                # prepare IDL script and add to users IDL runtime directory
                $toolStruct->{SCRIPT} = $toolScriptFile;
                $toolStruct->{IDL} = lc($toolIDLFile);
                prepareIDL($toolStruct,'tool');

                # run tool
                runIDL($toolStruct);
            } else {
                processToolMapping($query_instance, $tool_mappingfile,
                                   $mapFileIndex, $toolStruct);
            }
        }
    }
}

sub getToolName
{
    my ($toolowner, $toolname) = @_;
    my $toolfile;

    # check tool exists
    if ($toolowner eq "SYSTEM")
    {
        $toolfile = objectRealName($toolowner,'T',$toolname);
        if (! -d $toolfile) {
            abortQuery("Tool $toolowner\!$toolname not available");
        }
    }
    else
    {
        $toolfile = objectName($user,$toolowner,'T',$toolname);
        if (! -l $toolfile) {
            abortQuery("Tool $toolowner\!$toolname not available");
        }
    }
    $statusfile = joinName($toolfile,'status');
    if (! -e $statusfile)
    {
        abortQuery("No status file for tool $toolfile");
    }
    if (-s $statusfile == 0)
    {
        abortQuery("No information in status file for tool $toolfile");
    }
    if (checkStatus($statusfile) ne 'Y')
    {
        abortQuery("Tool $toolfile is not executable");
    }

    return $toolfile;
}

sub processDatasetList
{
    my ($datasetlist, $toolStruct) = @_;
    local ($ifile, $ivec, $infile);
    local (@inputDataFiles, @inputVectors);
    local ($toolnum, $tool, $toolname, $outputnum);

# find datasets
    @datasetlist = split(/;\s*/, $datasetlist);
    chomp @datasetlist;

    $ifile = 0;
    $ivec = 0;
    $infile = 0;
    while (@datasetlist)
    {
        ($datatype, $dataset) = split(/:/, shift(@datasetlist),2);

        # check for free input
        if ($dataset ne '?')
        {
            # process fixed input vectors and datasets
            # except where dataset is output from earlier tool
            if (index($dataset,'>') != -1)
            {
                ($toolnum,$tool) = split(/!/,$dataset,2);
                ($toolname,$outputnum) = split(/>/,$tool);
                $toolnameOnly = (split(/!/, $toolname))[1];
                if ($toolname ne '[this]')
                {
                    $dataname = $toolStruct->{QUERYNAME} . "_" .
                        $toolnameOnly . "_" . $toolnum . "_" . $outputnum;
                }
                else
                {
                    $dataname = $toolStruct->{QUERYNAME} . "_" .
                        $toolStruct->{CTOOLNAME} . "_" .
                        $toolnum . "_" . $outputnum;
                }
                $inputDataFiles[$ifile] = 'data';
                $inputDataDirs[$ifile] = objectName($user,$user,'D',$dataname);
                $ifile++;
            }
            else
            {
                foreach ($datatype)
                {
                    /^V/ && do # Vector
                    {
                        # add vector to IDL structure
                        $inputVectors[$ivec] = substr($dataset,1,length($dataset)-2);
                        $ivec++;

                        last;
                    };
                    /^D/ && do # Datafile
                    {
                        # Ultimately we shall just use the $dataset
                        # value as the name which will be the same as the
                        # identifier for the dataset in the LRD of the STPDF.
                        #
                        # For now, we shall use the AMPTE data explicitly


                        ($dataowner, $dataname) = split(/!/, $dataset);
                        if ($dataowner eq "SYSTEM")
                        {
                            $datafile = $dataset;
                            $datadir  = 'LRD';
                            $realDataFile = objectRealName($dataowner,'D',$dataname);
                        }
                        else
                        {
                            $datafile = 'data';
                            $datadir  = objectName($user,$dataowner,'D',$dataname);
                            $realDataFile = $datadir;
                        }
                        if (! -e $realDataFile)
                        {
                            abortQuery("Dataset $dataowner\!$dataname not available");
                        }

                        # if (! -e $datafile)
                        # {
                        #     abortQuery("Dataset $dataowner\!$dataname not available");
                        # }

                        # add data input file to IDL structure
                        $inputDataFiles[$ifile] = $datafile;
                        $inputDataDirs[$ifile]  = $datadir;
                        $ifile++;

                        last;
                    };
                }
            }
        }
        else
        {
            if ($datatype =~ /^D/) {
                $inputDataFiles[$ifile] = shift @ {$toolStruct->{INPUTFILES}};
                $inputDataDirs[$ifile]  = shift @ {$toolStruct->{INPUTDIRS}};
                $ifile++;
            } else {
                $inputVectors[$ivec] = shift @ {$toolStruct->{INPUTVECTORS}};
                $ivec++;
            }
        }
    }

    $toolStruct->{INPUTVECTORS} = [@inputVectors];
    $toolStruct->{INPUTFILES}   = [@inputDataFiles];
    $toolStruct->{INPUTDIRS}    = [@inputDataDirs];

    return $toolStruct;
}

sub prepareIDL {

    my ($toolStruct,$type) = @_;
    my ($fhIn,$line,@sedatFunctions,@sedatProcedures,$sfre,$spre);

#    @sedatFunctions  = 'write_log_record';
    @sedatProcedures = ('sedat_make_dmeta',
                        'sedat_make_meta',
                        'sedat_close_plot',
                        'sedat_close',
                        'sedat_open_plot',
                        'sedat_open_subset',
                        'sedat_open_read',
                        'sedat_open_write',
                        'sedat_read_record',
                        'sedat_read_value',
                        'sedat_write_record');
#    $sfre = join('|',@sedatFunctions);
    $spre = join('|',@sedatProcedures);

    # turn script into proper IDL
    # write header of IDL script (i.e. PRO )

    $fhIn = 'INSCRIPT';
    open $fhIn,$toolStruct->{SCRIPT};
    open OUTSCRIPT,">". joinName(${IDL_dir},$toolStruct->{IDL}).'.pro' or
        abortQuery("Can't open " . $toolStruct->{IDL} . "\n$!");

    if ($type eq 'tool') {
        print OUTSCRIPT "pro $toolStruct->{IDL}, cmn=cmn\n";
    }

    while (defined ($line = gatherLine($fhIn))) {
        # Add CMN parameter to PRO lines
        ($line =~ /^\s*pro /) && do {
            chomp $line;
            $line .= ", cmn=cmn\n";
        };
        # Add CMN parameter to FUNCTION lines
        ($line =~ /^\s*function /) && do {
            $line =~ s/\)\s*$/,cmn=cmn\)/;
        };
#        # Add CMN parameter to system functions
#        # This assumes at least one argument
#        @match = ($line =~ /($sfre)/);
#        if (@match > 0) {
#            $line =~ s/$match[0]\s*\(/$match[0]\(cmn=cmn,/;
#        };
        # Add CMN parameter to system procedures
        # This assumes at least one argument
        @match = ($line =~ /($spre)/);
        if (@match > 0) {
            $line =~ s/$match[0]\s*,/$match[0], cmn=cmn,/;
        }
        print OUTSCRIPT $line;
    }
    if ($type eq 'tool') {
        print OUTSCRIPT "\nend\n";
    }

    close $fhIn;
    close OUTSCRIPT;

}

sub gatherLine
{
    my ($fh) = @_;
    my ($line, $newline);

    $line = undef;
    while ($newline = <$fh>) {
#       $newline =~ s/;.*//;
        $line .= $newline;
        ($newline =~ /\$\s+$/) or last;
    }
    return $line;
}


sub runIDL
{
    my ($toolStruct) = @_;

    # create structure for IDL
    $odsPrefix = joinName(objectDir($toolStruct->{QUERYDIR},'D'),"${user}!");
    $dsDir     = joinName(objectDir(objectRealName($user,'U',$user),'D'),'');

    $struct    = $toolStruct->{QUERYNAME};
    $numInputFiles   = $#{$toolStruct->{INPUTFILES}}   + 1;
    $numInputValues  = $#{$toolStruct->{INPUTVECTORS}} + 1;
    $numOutputFiles  = $#{$toolStruct->{OUTPUTFILES}}  + 1;

    $arrsizeinf  = ($numInputFiles == 0)  ? 1 : $numInputFiles;
    $arrsizeoutf = ($numOutputFiles == 0) ? 1 : $numOutputFiles;
    $arrsizeinv  = ($numInputValues == 0) ? 1 : $numInputValues;

    $command
        = qq!$struct = {user:"$user",query_name:"",tool_name:"",ds_dir:"",!.
          qq!ods_prefix:"",query_dir:"",!.
          qq!idsf:strarr($arrsizeinf),idsd:strarr($arrsizeinf),!.
          qq!ivs:strarr($arrsizeinv),ods:strarr($arrsizeoutf),!.
          qq!idsid:lonarr($arrsizeinf),odsid:lonarr($arrsizeoutf)}!;
    IDLexecute("$command");

    IDLexecute(qq{$struct.query_name = "$struct"});
    IDLexecute(qq{$struct.tool_name = "$toolStruct->{TOOLNAME}"});
    IDLexecute(qq{$struct.ds_dir = "$dsDir"});
    IDLexecute(qq{$struct.ods_prefix = "$odsPrefix"});
    IDLexecute(qq{$struct.query_dir = "$toolStruct->{QUERYDIR}"});

    # Display the query input arguments
    IDLexecute("print, 'Query input arguments'");
    IDLexecute("print, '====================='");
    map {IDLexecute(qq{print, '$_'})} @query_args;
    IDLexecute("print,''");


    # input data files
    for ($infile=0; $infile<=$#{$toolStruct->{INPUTFILES}}; $infile++) {
        IDLexecute(qq{$struct.idsf[$infile] = "$toolStruct->{INPUTFILES}[$infile]"});
        IDLexecute(qq{$struct.idsd[$infile] = "$toolStruct->{INPUTDIRS}[$infile]"});
    }

    # input values
    for ($invec=0; $invec<=$#{$toolStruct->{INPUTVECTORS}}; $invec++) {
        IDLexecute(qq{$struct.ivs[$invec] = '$toolStruct->{INPUTVECTORS}[$invec]'});
    }

    # output data files
    for ($outfile=0; $outfile<=$#{$toolStruct->{OUTPUTFILES}}; $outfile++) {
        IDLexecute(qq{$struct.ods[$outfile] = "$toolStruct->{OUTPUTFILES}[$outfile]"});
    }

    # execute tool
    IDLexecute("$toolStruct->{IDL}, cmn=$struct");
    for ($i=0; $i<$numOutputFiles; $i++) {
        IDLexecute("s = call_external('libidl_stpdf.so','idl_stp_relcd',".
                   "${struct}.odsid[$i])");
    }
    for ($i=0; $i<$numInputFiles; $i++) {
        IDLexecute("s = call_external('libidl_stpdf.so','idl_stp_relcd',".
                   "${struct}.idsid[$i])");
    }
    IDLexecute("retall");
    IDLexecute("delvar, $struct");

}

sub compileIDL
{
    my ($toolStruct) = @_;

    IDLexecute(".compile $toolStruct->{IDL}.pro");
}

sub IDLexecute
{
    my ($IDLcommand) = @_;

    # Add the command to the query instance commands file

    print QUERYCMDS $IDLcommand,"\n";
}

sub abortQuery {
    ($msg) = @_;

    print STDOUT $msg;
    exit(1);
}
