<?php
/**
* Library functions for the ODI database. 
*
* All functions that are related to the operation of the ODI database are defined in this library.
* Functions for parsing CDF and text files are defined in {@link parsefile.php} and
* {@link dirscan.php}.
*
* <pre>
* <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
* <>          ___  ___  __ _                                        <>
* <>         / _ \/ __|/ _` |               european space agency   <>
* <>        |  __/\__ \ (_| |                                       <>
* <>         \___/\___/\__,_|            agence spatiale europeenne <>
* <>                                                                <>
* <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
* </pre>
* The IPR of this software belongs to the European Space Agency. 
* The usage is at the discretion of the Agency and licenses for its 
* use should be addressed to the Contracts Dept. at ESTEC: 
* <pre>
* European Space Agency ESA/ESTEC 
* Contracts Department 
* Postbus 299 
* 2200 AG Noorwijk 
* The Netherlands 
* 
* WEB:         http://www.esa.int/ 
* Phone:         +31 71 565-6565 
* </pre>
* The software was developed under the ESA/ESTEC contract 21964/08/NL/AT by: 
* <pre>
* Peter Wintoft, Swedish Institute of Space Physics and 
* Daniel Heynderickx, DH Consultancy
* </pre>
* 
* Version number: 1.0 
*
* Release Date: 2 December, 2009 
*
* @version $Id: odi.library.php 314 2010-06-15 10:46:15Z daniel $
*
* @package ODI
* @subpackage Library
*/

//require_once($_SERVER["ODI_HOME"]."/parsers/dirscan.php");

//require_once($_SERVER["ODI_HOME"]."/parsers/parsefile.php");

// Read the revision info.
$fn=$_SERVER["ODI_HOME"]."/revision.txt";
if(file_exists($fn)) {
    define("REVISION", trim(file_get_contents($fn)));
}
else {
    define("REVISION", "Unknown");
}

/**
* The line-feed depends on the OS.
*/
// Set the correct line feed depending on OS.
if (substr(PHP_OS, 0, 3) == "WIN")
  define("LF", "\r\n");
else
  define("LF", "\n");

/**
* The date-number of the CDF 1970 epoch.
*/
define("CDF1970","62167219200000.0");

/**
* Convert CDF Epoch to UTC date string of the form YYYY-MM-DD hh:mm:ss
*
* @param double $cdf_epoch The CDF EPOCH
* @return array The date and milliseconds
*/
function cdfepoch_to_date($cdf_epoch) {
    $epoch = $cdf_epoch - 62167219200000.0;
    $time = (int) ($epoch/1000.0);
    $ms = (int) ($epoch - $time * 1000.0);
    if ($ms < 0)
    {
      $time -= 1;
      $ms += 1000;
    }
    $date = gmdate("Y-m-d H:i:s", $time);
    return array($date, $ms);
}

/**
* Check if dataset is listed in table dataset and return the dataset_id
*
* @param mysqli   $link          A MySQL database connection
* @param string   $tablename     The ODI name of the dataset
* @param resource $lhandle       A file pointer to the log file.
* @return int The dataset_id
*/
function checkTable($link, $tablename, $lhandle)
{
    $tableid = FALSE;
    
    $query = "SELECT * FROM dataset WHERE " .
             "name='" . $tablename . "'";
    $result = mysqli_query($link, $query);
    if (mysqli_num_rows($result) > 0)
    {
        $row = mysqli_fetch_array($result, MYSQLI_ASSOC);
        $tableid = $row["id"];
        mysqli_free_result($result);
    }
    else
        fprintf($lhandle, "  Table " . $tablename . " not found" . LF);
    
    return $tableid;
}

/**
* Read and clean the skeleton file to enable further parsing.
*
* @param string $sk_file_name The skeleton file name.
* @return array An array with the contents of the skeleton file.
*/
function clean_skeleton($sk_file_name) {
    // Read in all rows and make the data consistent.
    $sk_file=fopen($sk_file_name,"r");
    $row_num=0;
    while($row=fgets($sk_file)) {
        // Remove leading and trailing spaces.
        $row=trim($row);
        // Skip comment lines and blank lines.
        if(strlen($row)>0 & preg_match("/^!/",$row)==0) {
            // If the data is split over several rows then merge.
            // First check if it just one string enclosed in "".
            if(preg_match('/".+"$/',$row)) {
                // Then check if the line following is a continuation.
                $row_next=trim(fgets($sk_file));
                // Is it a global attribute continuation or a variable attribute continuation?
                if(preg_match('/\d+:.*/',$row_next) | preg_match('/\w+\s+.+/',$row_next))
                    $row .= " ".$row_next;
                // Else treat it as a new row.
                else {
                    $row_all[$row_num] = $row;
                    $row_num++;
                    $row = $row_next;
                }
            }            
            // If row ends with "-" then read next row and append to current row.
            while(preg_match('/(.*)"\s+\-$/',$row,$match_1)) {
                $row_next=trim(fgets($sk_file));
                preg_match('/"(.+)/',$row_next,$match_2);
                $row = $match_1[1].$match_2[1];
            }
            // Add rows to row_all.
            $row_all[$row_num]=$row;
            $row_num++;
        }
    }    
    fclose($sk_file);
    return $row_all;
}

/**
* Connect to the database using an array with connection parameters.
*
* @param array $ca An array with connection parameters (see get_conn_array)
* @return array MySQL connection.
*/
function connect_to_database_array($ca) {
    $conn = mysqli_init();
    if (!$conn) die("Failed to initialize database connection");
    if (!$conn->options(MYSQLI_OPT_LOCAL_INFILE, 1)) die("Failed to set LOAD LOCAL option");
    if (!$conn->real_connect(
        $ca["host"],$ca["user"],$ca["password"],$ca["database"],$ca["port"],$ca["socket"])) die("Unable to connect to database");
    return $conn;
}

/**
* Create a column name with data type to be used in creating dataset_* table.
*
* @param string $name A string with the base part of the name.
* @param array  $num  An array with numbers.
* @param string $type The MySQL data type.
* @return string The column name.
*/
function create_column_name($name,$num,$type) {
    $col_name = $name;
    if(count($num)>0) {
        foreach($num as $n) {
            $col_name .= "_".$n;
        }
    }
    $col_name .= " ".$type.",";
    return $col_name;
}

/**
* Create a "dataset_*" table.
*
* The column names are constructed from "data" (CDF VAR_TYPE="data") variable names and
* the NRV index if it is larger than 0.
* The tables always have the columns cdf_epoch, epoch, millisec, and dataset_file_id.
* Updated version with improved query syntax.
*
* @param mysqli $conn         A MySQL database connection.
* @param int    $dataset_id   The id of a selected dataset.
* @param string $dataset_name The name of the table holding the data.
*/
function create_dataset_table($conn,$dataset_id,$dataset_name) {
    
    // Get the dataset id and majority.
    $q = 'SELECT majority FROM dataset WHERE id=' . $dataset_id;
    if (!($result=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
    $row = $result->fetch_assoc();
    $first_dim_slow = TRUE;         // If ROW_MAJOR then the first dimension changes the slowest.
    if(strncasecmp($row["majority"],"col",3) == 0) {
        $first_dim_slow = FALSE;    // If COLUMN_MAJOR then the first dimension changes the fastest.
    }
    
    // Create columns for all variables that are record variant except Epoch.
    $q  = 'SELECT vm.name AS column_name,';
    $q .= ' vm.sizes AS size,';
    $q .= ' dt.mysql_type AS data_type,';
    $q .= ' vm.dimension_variances AS dim_var';
    $q .= ' FROM variable_metadata AS vm';
    $q .= ' LEFT JOIN data_type AS dt ON (vm.data_type_id=dt.id)';
    $q .= ' WHERE vm.dataset_id=' . $dataset_id;
    $q .= ' AND vm.record_variance="T"';
    $q .= ' AND STRCMP(LOWER(vm.name),"epoch")!=0';
    if (!($result=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
    $q = 'CREATE TABLE '.$dataset_name.' ( ';
    $q .= 'cdf_epoch DOUBLE NOT NULL,';
    $q .= 'epoch DATETIME,';
    $q .= 'millisec SMALLINT,';
    while($row=$result->fetch_assoc()) {
        // Get the name part and the data type to form the query.
        $col_name = s_to_us(strtolower($row["column_name"]));
        $data_type = $row["data_type"];
        // If "size" is set then it is a multidimensional variable else it is a scalar.
        if($row["size"]) {
            $size=explode(",",$row["size"]);

            if($first_dim_slow) $size = array_reverse($size);
            
            $d = count($size);
            $n = array_product($size);
            $a = array_fill(0,$d,1);
            $postfix = array();
            $postfix[] = $a;
            for($i=1;$i<$n;$i++) {
                $a[0]++;
                if($a[0]>$size[0]) {
                    $a[0] = 1;
                    $j = 1;
                    $a[$j]++;
                    while($a[$j]>$size[$j]) {
                        $a[$j] = 1;
                        $j++;
                        $a[$j]++;
                    }
                }
                if($first_dim_slow) {
                    $postfix[] = array_reverse($a);
                }
                else {
                    $postfix[] = $a;
                }
            }
            
            foreach($postfix as $p) {
                $q .= create_column_name($col_name,$p,$data_type);
            }

        }
        else {
            $q .= create_column_name($col_name,null,$data_type);
        }
    }
    $q .= 'dataset_file_id INT UNSIGNED,';
    $q .= 'PRIMARY KEY(cdf_epoch),';
    $q .= 'KEY(dataset_file_id)';
    $q .= ') ENGINE=INNODB';
    $q .= ' COMMENT='.quote(REVISION);
    if (!($conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
    $q = 'ALTER TABLE '.$dataset_name.' ADD FOREIGN KEY (dataset_file_id) ';
    $q .= 'REFERENCES dataset_file (id) ON DELETE CASCADE';
    if (!($conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
}

/**
* Update table dataset with start and end epoch and modification date.
*
* @param mysqli $link      A MySQL database connection
* @param string $tablename The name of the dataset_* table.
* @param int    $tableid   The id in dataset for the dataset_* table.
*/
function dataset_update($link, $tablename, $tableid)
{
    $query = "SELECT COUNT(cdf_epoch), MIN(cdf_epoch), MAX(cdf_epoch) FROM " . $tablename;
    $result = mysqli_query($link, $query);
    if (mysqli_num_rows($result) > 0)
    {
        $row = mysqli_fetch_array($result, MYSQLI_NUM);
        mysqli_free_result($result);
        $query = "UPDATE dataset SET numrecs=" . $row[0] . ", epoch_start=" . $row[1] .
            ", epoch_end=" . $row[2] . ", modification_date=NOW() " .
            "WHERE id=" . $tableid . " LIMIT 1";
        mysqli_query($link, $query);
    }
}

/**
* A function to convert a date into CDF EPOCH.
*
* @param string $date     The date in YYYY-MM-DD hh:mm:ss format
* @param double $millisec The millisecond part.
* @return double The CDF EPOCH
*/
function date_to_cdfepoch($date, $msec) {
    $cdf_1970 = 62167219200000.0;

    $epoch = 1000.0 * gmmktime(substr($date, 11, 2), substr($date, 14, 2), substr($date, 17, 2),
                               substr($date, 5, 2), substr($date, 8, 2), substr($date, 0, 4));
    $epoch += $cdf_1970 + $msec;

    return $epoch;
}
  
/**
* Delete the "dataset_*" table together with all related entries.
*
* The function first drops the table holding the data (i.e. the dataset_* table).
* Then the entry in the dataset table is deleted and all associated entries
* in the other tables are automatically deleted.
*
* @param mysqli $conn         A MySQL database connection.
* @param string $dataset_name The name of the table holding the data.
*/
function delete_dataset($conn,$dataset_name) {
    $q = 'DROP TABLE '.no_quote($dataset_name);
    if (!($conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
    $q = 'DELETE FROM dataset WHERE name='.quote($dataset_name);
    if (!($conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
}

/**
* Recursive directory scan.
*
* @param string $basedir     The base directory to scan.
* @param string $filepattern The pattern to match against the file names.
*/
function dirscan($basedir, $filepattern)
{
    $dir_iterator = new RecursiveDirectoryIterator($basedir);
    $iterator = new RecursiveIteratorIterator($dir_iterator, RecursiveIteratorIterator::SELF_FIRST);

    $dfiles = array();
    foreach ($iterator as $file)
    {
        $fget = TRUE;
        if ($filepattern != "%")
        {
            $parse = explode("%", trim($filepattern, "%"));
// If the filepattern does not start with a wildcard, match from the start of the filename
            if (substr($filepattern, 0, 1) != "%")
                $fget = $fget && (substr($file->getFilename(), 0, strlen($parse[0])) == $parse[0]);
            for ($i = 0; $i < count($parse); $i++)
                $fget = $fget && (strpos($file->getFilename(), $parse[$i]) !== FALSE);
        }
        if ($fget === TRUE)
        {
            $date = gmdate("Y/m/d h:i:s", $file->getMTime());
            $dfiles[] = array($file->getFilename(), $date, $file->getPathname());
        }
    }
    $tmpfiles = array();
    foreach ($dfiles as $key => $row)
        $tmpfiles[$key] = $row[2];
    array_multisort($tmpfiles, SORT_ASC, $dfiles);

    return $dfiles;
}

/**
* Check if dataset name exists in dataset table.
*
* @param mysqli $conn         A MySQL database connection.
* @param string $dataset_name The name of the table holding the data.
* @return bool True if the dataset exists.
*/
function exists_dataset($conn,$dataset_name) {
    $q = 'SELECT COUNT(id) AS num FROM dataset WHERE name='.quote($dataset_name);
    if (!($res=$conn->query($q))) printf("Query: %s".LF."Error: %s".LF,$q,$conn->error);
    $row=$res->fetch_assoc();
    return ($row["num"]>0);
}

/**
* Connection parameters for a given user level.
*
* An array is returned with the ODI database parameters
* based on the environment variables and the user level.
*
* @param integer $user_level The user level
* @return array Array with connection parameters
*/
function get_conn_array($user_level) {

    // List with parameters to be used to connect to the database.
    // If the environment variable is not required it is marked as false.
    $param=array(
        "host"=>array("ODI_HOST",false),
        "user"=>array("ODI_USER_".$user_level,true),
        "password"=>array("ODI_PW_".$user_level,true),
        "port"=>array("ODI_PORT",false),
        "socket"=>array("ODI_SOCKET",false),
        "database"=>array("ODI_DB",true)
    );

    $ca=array();
    
    foreach($param as $key=>$val) {
        list($env_var,$req) = $val;
        if($ev=getenv($env_var)) {
            $ca[$key] = $ev;
        }
        else {
            if($req) {
                die("Environment variable ".$env_var." must be set!".LF);
            }
            $ca[$key] = null;
        }
    }
    
   return $ca;
}

/**
* A support function to get the command line arguments when a script is executed
* on the command line.
*
* The first argument ($argv) is an array with the actual arguments given on
* the command line, including the script name itself.
*
* The second argument ($arg_array) is an array where the values will become the
* keys to the returned array.
*
* Example:
*
*  Executing a script on command line like
*  <code>
*    show_metadata.php DST
*  </code>
*  uses
*  <code>
*    $arg_array = array("dataset");
*    $args = get_cl_arguments($argv,$arg_array);
*  </code>
*  where the array $args becomes
*  <code>
*    $args = array ( "dataset" => "DST" )
*  </code>
*
* @param array $argv 
* @param array $arg_array
* @return array 
*/
function get_cl_arguments($argv,$arg_array) {
    // Check number of input arguments.
    if (count($argv)<(sizeof($arg_array)+1)) {
        fwrite(STDOUT,"Error: Too few input arguments.\n");
        fwrite(STDOUT,"Usage: ".$argv[0]);
        foreach($arg_array as $a)
            fwrite(STDOUT," <".$a.">");
        fwrite(STDOUT,"\n");
        exit(0);
    }

    // Input parameters.
    array_shift($argv); #First argument is the script name.
    foreach($arg_array as $a)
        $args[$a]=array_shift($argv);
    return $args;
}

/**
* Get the data and field names for a set of variables for a dataset.
*
* @param mysqli $conn A MySQL connection.
* @param string $dataset The SEDAT name of the dataset.
* @param double $epoch_start_cdf Start epoch in CDF epoch format.
* @param double $epoch_end_cdf End epoch in CDF epoch format.
* @param int $max_num_rec Maximum number of records to be read.
* @param array $meta An array with variable metadata (name,dims,sizes).
* @return array An array with a MySQL result set and field name for each variable.
*/
//function get_data_and_fields($conn,$dataset,$epoch_start_cdf,$epoch_end_cdf,$max_num_rec,$meta) {
function get_data_and_fields($conn,$dataset,$epoch_start_cdf,$epoch_end_cdf,$meta) {

    $q = "SELECT name FROM dataset WHERE sedat_name=".quote($dataset);
    if (!($res=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
    $row = $res->fetch_row();
    $dataset = $row[0];

    $field_name = array();
    $q = "SELECT ";
    foreach($meta as $m) {
        if($m["dim_var"]=="F") {
            $m["dims"] = 0;
        }
        if(!strcasecmp("epoch",$m["name"])==TRUE) {
            $q.= "concat(epoch,'.',lpad(millisec,3,'0')),";
            $field_name[] = "Epoch";
        }
        else {
            switch ($m["dims"]) {
                case 0:
                    $q.= s_to_us(strtolower($m["name"])).",";
                    $field_name[] = $m["name"];
                    break;
                case 1:
                    $nrv_values = get_nrv_values($conn,$dataset,$m["name"]);
                    for($i=1;$i<=$m["sizes"];$i++) {
                        $q.= s_to_us(strtolower($m["name"]))."_".$i.",";
                        $field_name[] = $m["name"]." (".$nrv_values[$i-1].")";
                    }
                    break;
                case 2:
                    $nrv_values = get_nrv_values($conn,$dataset,$m["name"]);
                    list($i_max,$j_max) = explode($m["sizes"],",");
                    $k=0;
                    for($i=1;$i<=$i_max;$i++) {
                        for($j=1;$j<=$j_max;$j++) {
                            $q.= s_to_us(strtolower($m["name"]))."_".$i."_".$j.",";
                            $field_name[] = $m["name"]." (".$nrv_values[$k++].")";
                        }
                    }
                    break;
            }
        }
    }
    $q = trim($q,",");
    $q.= " FROM ".$dataset;
    $q.= " WHERE cdf_epoch>=".sprintf("%f",$epoch_start_cdf);
    $q.= " AND cdf_epoch<=".sprintf("%f",$epoch_end_cdf);
    $q.= " ORDER BY cdf_epoch";
//    $q.= " LIMIT ".no_quote($max_num_rec);
        
    if (!($res=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);

    return array($res,$field_name);
}

/**
* Get the NRV values for a variable.
* If the LABL_PTR_? attribute is not defined the returned values will just be
* an array of number.
*
* @param mysqli $conn          A MySQL database connection
* @param string $dataset       The ODI name of the dataset
* @param string $variable_name The variable name
* @return array $nrv_values    The NRV values in an indexed array
*/
function get_nrv_values($conn,$dataset,$variable_name) {
    // Get the sizes of the variable.
    $q = "SELECT vm.id AS id, vm.sizes AS sizes";
    $q.= " FROM variable_metadata AS vm";
    $q.= " LEFT JOIN dataset AS ds ON vm.dataset_id=ds.id";
    $q.= " WHERE ds.name=".quote($dataset);
    $q.= " AND vm.name=".quote($variable_name);
    if (!($res=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
    $var_id=0;
    $n=0;
    if($row=$res->fetch_assoc()) {
        $var_id=$row["id"];
        $sizes = explode(",",$row["sizes"]);
        $n = array_product($sizes);
    }
    $nrv_values = array();
    // Get the NRV values.
    $q = "SELECT va.value AS nrv_var";
    $q.= " FROM variable_attribute AS va";
    $q.= " WHERE va.variable_metadata_id=".$var_id;
    $q.= " AND va.name='labl_ptr_1'";
    if (!($res=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
    if($row=$res->fetch_assoc()) {
        $q = "SELECT nrv.value FROM variable_nrv AS nrv";
        $q.= " LEFT JOIN variable_metadata AS vm ON nrv.variable_metadata_id=vm.id";
        $q.= " WHERE vm.name=".quote($row["nrv_var"]);
        if (!($res=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
        while($row=$res->fetch_row()) {
            $nrv_values[] = trim($row[0]);
        }
    }
    if(count($nrv_values)<$n) {
        $nrv_values = range(1,$n);
    }
    return $nrv_values;
}

/**
* Get the metadata for a list of variables.
*
* @param mysqli $conn A MySQL connection.
* @param string $dataset The SEDAT name of the dataset.
* @param array $var An array with variable names.
* @return array An array with metadata for each variable.
*/
function get_variable_metadata($conn,$dataset,$var) {
    $meta = array();
    foreach($var as $v) {
        $q = "SELECT vm.name,vm.dims,vm.sizes,vm.dimension_variances AS dim_var FROM variable_metadata AS vm";
        $q.= " LEFT JOIN dataset ON vm.dataset_id=dataset.id";
        $q.= " WHERE dataset.sedat_name=".quote($dataset);
        $q.= " AND vm.name=".quote($v);
        if (!($res=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
        if($row=$res->fetch_assoc()) {
            $meta[] = $row;
        }
        else {
            printf("Query: %s\nError: No data!\n",$q);
        }
    }
    return $meta;
}


/**
* Load temporary files load.tmp into dataset_* and update dataset and dataset_files.
* 
* @param mysqli   $link    A MySQL database connection
* @param array    $dfiles Multi-dim array with file names, dates, and complete path to file.
* @param array    $tfiles Multi-dim array with id, file names, and dates of file already ingested.
* @param resource $lHandle A file pointer to the log file.
* @param string   $tablename The name of the dataset_* table.
* @param int      $tableid The id of the dataset.
* @param string   $platform The platform name.
* @param string   $instrument The instrument name.
* @param string   $sk_file The name of the skeleton file
* @return int The number of files loaded.
*/
function getFiles($link, $dfiles, $tfiles, $lHandle, $tablename, $tableid, $platform, $instrument, $sk_file,
                    $parser)
{
    $nf = 0;
    foreach ($dfiles as $dfile)
    {
        $fid = FALSE;
    
        $filename = $dfile[0];
        $filedate = $dfile[1];
        $mtime = strtotime($filedate);
        $rfname = $dfile[2];
    
        $getFile = false;
        $key = FALSE;
        if (count($tfiles) > 0) $key = array_search($filename, $tfiles["dfname"]);
        if ($key !== FALSE)
        {
            $fid = $tfiles["id"][$key];
            $sfiledate = $tfiles["fdate"][$key];
            if ($mtime > strtotime($sfiledate)) $getFile = true;
        }
        else
            $getFile = true;
    
        if ($getFile)
        {
            $sqlerror = FALSE;
            fprintf($lHandle, "%s", $rfname . "  ");
            if ($fid !== FALSE)
            {
                $query = "UPDATE dataset_file SET filedate='" . $filedate .
                         "',ingestion_date=NOW() " .
                         "WHERE id=" . $fid . " LIMIT 1";
                mysqli_query($link, $query);
                if (mysqli_affected_rows($link) > 0)
                    $status = "updated";
                else
                {
                    $status = "update error";
                    $sqlerror = mysqli_error($link);
                }
            }
            else
            {
                $query = "INSERT INTO dataset_file VALUES(DEFAULT,'" .
                        $filename . "'," . $tableid . ",'" . $filedate .
                        "',NOW())";
                mysqli_query($link, $query);
                if (mysqli_affected_rows($link) > 0)
                {
                    $fid = mysqli_insert_id($link);
                    $status = "loaded";
                }
                else
                {
                    $status = "insert error";
                    $sqlerror = mysqli_error($link);
                }
            }
            if (strpos($status, "error") === FALSE)
            {
                $result = $parser($rfname, $fid, $sk_file);
                if ($result)
                {
                    $query = "LOAD DATA LOCAL INFILE 'load.tmp' " .
                             "REPLACE INTO TABLE `" . $tablename .
                             "` FIELDS TERMINATED BY ',' " .
                             "OPTIONALLY ENCLOSED BY '\"' " .
                             "LINES TERMINATED BY '" . LF . "'";
                    mysqli_query($link, $query);
                    if (mysqli_affected_rows($link) <= 0)
                    {
                        $status = "load error";
                        $sqlerror = mysqli_error($link);
                    }
                    else
                        $nf++;
                    @unlink("load.tmp");
                }
                else
                    $status = "read error";
            }
            if ((strpos($status, "error") !== FALSE) && ($fid !== FALSE))
            {
                $query = "DELETE FROM dataset_file WHERE id=" . $fid . " LIMIT 1";
                mysqli_query($link, $query);
            }
            fprintf($lHandle, "%s", $status);
            if ($sqlerror !== FALSE)
                fprintf($lHandle, "  %s", $sqlerror);
            fprintf($lHandle, "%s", LF);
        }
    }
    
    return $nf;
}

/**
* Construct a list of files to be ingested, convert them to temporary file and ingest into the
* database.
*
* @param mysqli $link    A MySQL database connection
* @param string $tablename The name of the dataset_* table.
* @param string $platform The platform name.
* @param string $instrument The instrument name.
* @param string $filepattern The file pattern to match against the raw data files.
* @param string $sk_file The name of the skeleton file
* @return int The number of files ingested.
*/
function ingest_data($link, $tablename, $basedir, $platform, $instrument, $filepattern, $sk_file,
                        $scanner, $parser)
{
    $nf = 0;
    $lname = $instrument . ".log";
    $lhandle = fopen($lname, "a");
    fprintf($lhandle, LF . "--" . LF . "-- " . date("d/m/Y H:i:s") . LF .
              "--" . LF . LF);

    $tableid = checkTable($link, $tablename, $lhandle);

    if ($tableid !== FALSE)
    {
        $tfiles = tfileList($link, $tableid);
//        $dfiles = rfileList($platform, $instrument, $basedir, $filepattern);
        $dfiles = $scanner($basedir, $filepattern);
        if (count($dfiles) > 0)
            $nf = getFiles($link, $dfiles, $tfiles, $lhandle, $tablename, $tableid, $platform, $instrument, $sk_file,
                            $parser);
        if ($nf > 0) dataset_update($link, $tablename, $tableid);
    }

    fclose($lhandle);

    return $nf;
}

/**
* Insert a row in the dataset table.
*
* @param mysqli $conn         A MySQL database connection.
* @param string $dataset_name The name of the table holding the data.
*/
function insert_dataset_entry($conn,$dataset_name) {
    $dataset_id = FALSE;
    $q = 'INSERT INTO dataset (name,creation_date)';
    $q .= ' VALUES ('.quote($dataset_name).',NOW())';
    if (!($conn->query($q)))
      printf("Query: %s".LF."Error: %s".LF,$q,$conn->error);
    else
      $dataset_id = $conn->insert_id;

    return $dataset_id;
}    

/**
* Insert a row in the dataset_metadata table.
*
* The <code>$meta</code> is an array with the following keys:
* <code>
* "dataset", "attribute", "entry_number", "value", "data_type"
* </code>
*
* @param mysqli $conn A MySQL database connection.
* @param array  $meta The name of the table holding the data.
*/
function insert_dataset_metadata_entry($conn,$meta) {
    $q = 'INSERT IGNORE INTO dataset_metadata';
    $q .= ' (dataset_id,attribute,data_type_id,entry_number,value)';
    $q .= ' SELECT ' . $meta["dataset"] . ',';
    $q .= quote($meta["attribute"]).',data_type.id,'.$meta["entry_number"].',';
    $q .= quote($meta["value"]);
    $q .= ' FROM data_type ';
    $q .= ' WHERE data_type.cdf_type='.quote($meta["data_type"]);
    if (!($conn->query($q))) printf("Query: %s".LF."Error: %s".LF,$q,$conn->error);
}

/**
* Insert a row in the variable_attribute table.
*
* The variable attributes are in the array <code>$metadata</code> and has the following keys:
* <code>
* "attribute"
* "value"
* "variable_metadata"
* "data_type"
* </code>
*
* @param mysqli $conn       A MySQL database connection.
* @param array  $metadata   The variable attributes.
* @param int    $dataset_id The dataset id.
*/
function insert_variable_attribute_entry($conn,$metadata,$dataset_id) {
    $q = 'INSERT IGNORE INTO variable_attribute ';
    $q .= '(name,variable_metadata_id,data_type_id,value) ';
    $q .= 'SELECT '.quote($metadata["attribute"]).',variable_metadata.id';
    $q .= ',data_type.id,'.quote($metadata["value"]);
    $q .= ' FROM variable_metadata, data_type';
    $q .= ' WHERE variable_metadata.name='.quote($metadata["variable_metadata"]);
    $q .= ' AND variable_metadata.dataset_id=' . $dataset_id;
    $q .= ' AND data_type.cdf_type='.quote($metadata["data_type"]);
    if (!($conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
}

/**
* Insert a row in the variable_metadata table.
*
* The variable metadata are in the array <code>$metadata</code> and has the following keys:
* <code>
* "name"
* "dataset"
* "number_elements"
* "dims"
* "sizes"
* "record_variance"
* "dimension_variances"
* "data_type"
* </code>
*
* @param mysqli $conn    A MySQL database connection.
* @param array  $metdata The variable metadata.
*/
function insert_variable_metadata_entry($conn,$metadata) {
    $q = 'INSERT IGNORE INTO variable_metadata ';
    $q .= '(name,dataset_id,data_type_id,number_elements,dims,sizes,record_variance,dimension_variances) ';
    $q .= 'SELECT ';
    $q .= '"'.$metadata["name"].'",';
    $q .= '"'.$metadata["dataset"].'",';
    $q .= 'data_type.id,';
    $q .= $metadata["number_elements"].',';
    $q .= $metadata["dims"].',';
    $q .= '"'.$metadata["sizes"].'",';
    $q .= '"'.$metadata["record_variance"].'",';
    $q .= '"'.$metadata["dimension_variances"].'" ';
    $q .= 'FROM data_type ';
    $q .= 'WHERE data_type.cdf_type="'.$metadata["data_type"].'"';
    if (!($conn->query($q)))
        printf("Query: %s\nError: %s\n",$q,$conn->error);
}

/**
* Insert a row in the nrv_value table.
*
* @param mysqli $conn          A MySQL database connection.
* @param string $variable_name The name of the variable.
* @param int    $nrv_index     The index of an element in the NRV variable.
* @param string $nrv_pos       The positional index as given in the skeleton file.
* @param string $nrv_value     The value of the NRV variable at the given index.
* @param string $dataset_id    The id of the dataset.
*/
function insert_variable_nrv_entry($conn,$variable_name,$nrv_index,$nrv_pos,$nrv_value,$dataset_id) {
    $q = 'INSERT IGNORE INTO variable_nrv (variable_metadata_id,ind,pos,value) ';
    $q .= 'SELECT variable_metadata.id,'.$nrv_index.','.quote($nrv_pos).','.quote($nrv_value);
    $q .= ' FROM variable_metadata WHERE';
    $q .= ' variable_metadata.name='.quote($variable_name);
    $q .= ' AND variable_metadata.dataset_id=' . $dataset_id;
    if (!($conn->query($q))) printf("Query: %s".LF."Error: %s".LF,$q,$conn->error);
}

/**
* A function to remove quotes (").
*
* @param string $s A string, with or without quotes
* @return string A string without quotes
*/
function no_quote($s) {
    return trim($s,'\'"');
}

/**
* A function for reading ODI data into an array.
*
* @param string $dataset The SEDAT name of the dataset.
* @param string $epoch_start The start epoch in format "YYYY-MM-DD hh:mm:ss.millisec".
* @param string $epoch_end The end epoch in format "YYYY-MM-DD hh:mm:ss.millisec".
* @param int $max_num_rec The maximum number of records to be read.
* @param string $var The name of a variable to be read.
* @param string $var,... Unlimited optional variables to be read.
* @return array A 2-dimensional array with data. The record number increases along the first dimension, and the variables along the second dimension.
*/
function odi_read_data() {

    $arg_list = func_get_args();
    $dataset = array_shift($arg_list);
    $epoch_start = array_shift($arg_list);
    $epoch_end = array_shift($arg_list);
//    $max_num_rec = array_shift($arg_list);
    $var = $arg_list;

    $data = array();
    
    // Connect to the database at user level 3.
    $conn_array=get_conn_array(3);
    $conn=connect_to_database_array($conn_array);

    // Convert epoch range to CDF epoch.
    list($date,$ms) = explode(".",$epoch_start);
    $epoch_start_cdf = date_to_cdfepoch($date,$ms);
    list($date,$ms) = explode(".",$epoch_end);
    $epoch_end_cdf   = date_to_cdfepoch($date,$ms);

    // Get the metadata for each variable.
    $meta = get_variable_metadata($conn,$dataset,$var);
    
    // Get the data and field names.
//    list($res,$field_name) = get_data_and_fields($conn,$dataset,$epoch_start_cdf,$epoch_end_cdf,$max_num_rec,$meta);
    list($res,$field_name) = get_data_and_fields($conn,$dataset,$epoch_start_cdf,$epoch_end_cdf,$meta);

    while($row=$res->fetch_row()) {
        $data[] = array_combine($field_name,$row);
    }
    
    return $data;
}

/**
* A function to ensure that a string is enclosed in quotes (").
*
* @param string $s A string, with or without quotes
* @return string A string with quotes
*/
function quote($s) {
    return '"'.trim($s,'"').'"';
}

/**
* A function to replace non-word characters with underscore "_".
*
* @param string $s A string with possible non-word characters
* @return string A string without possible underscores
*/
function s_to_us($s) {
    return trim(preg_replace('/\W+/','_',$s),'_');
}

/**
* A function to create the ODI metadata from a CDF skeleton table.
*
* The metadata are stored into the ODI tables dataset, dataset_metadata,
* variable_metadata, and variable_attribute.
* @param mysqli $conn             A MySQL database connection.
* @param string $dataset_name     The name of the dataset (e.g. dataset_index_Dst).
* @param string $platform         The platform (e.g. index, XMM, ...).
* @param string $platform_type    The platform type (e.g. virtual, satellite, ...).
* @param string $instrument       The instrument (e.g. Dst, ERMD, ...).
* @param string $sk_file_name     The skeleton file name.
* @param string $availability     Availability: public or private.
* @param string $sedat_name       The SEDAT name.
* @param string $sedat_regions    The SEDAT region code.
* @param string $sedat_particles  The SEDAT particle code.
* @param string $sedat_emin       The SEDAT minimum energy.
* @param string $sedat_emax       The SEDAT maximum energy.
*/
function skt2odi($conn, $dataset_name, $platform, $platform_type, $instrument, $sk_file_name, $availability,
                 $sedat_name, $sedat_regions, $sedat_particles, $sedat_emin, $sedat_emax)
{
    
    $test_mode = FALSE;
    
    if($test_mode) {
        fwrite(STDOUT,"===========================================\n");
        fwrite(STDOUT,"Test mode.\n");
        fwrite(STDOUT,"Printing contents of cleaned skeleton file.\n");
        fwrite(STDOUT,"-------------------------------------------\n");
    }
    
    // Clean the skeleton file and put into an array.
    $row_all = clean_skeleton($sk_file_name);
    
    if($test_mode) {
        fwrite(STDOUT,"-------------------------------------------\n");
        fwrite(STDOUT,"Finished printing skeleton file.\n\n\n\n\n\n\n\n");
        fwrite(STDOUT,"===========================================\n");
        fwrite(STDOUT,"Listing interpreted skeleton file.\n");
        fwrite(STDOUT,"-------------------------------------------\n");
    }

    // Also check if the "dataset_*" table already exists, then delete it.
    // This can happen even when there is no entry in "dataset" table.
    if(!$test_mode) {
        $q = 'SHOW TABLES LIKE '.quote($dataset_name);
        if (!($res=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
        $row=$res->fetch_assoc();
        if($row) {
            fwrite(STDOUT,$dataset_name." exists.  Deleting ...\n");
            $q = 'DROP TABLE '.$dataset_name;
            if (!($res=$conn->query($q))) exit("Query: %s\nError: %s\n".$q.$conn->error);
        }
    
        // Insert an entry into the dataset table.
        fwrite(STDOUT,"Creating ".$dataset_name." and metadata ...\n");
        $dataset_id = insert_dataset_entry($conn,$dataset_name);
    
// Add the availability value
        $meta = array("dataset" => $dataset_id,
                      "attribute" => "availability",
                      "value" => $availability);
        update_dataset_entry($conn, $meta);

        // Add the SEDAT specific metadata
        $meta["attribute"] = "sedat_name";
        $meta["value"] = $sedat_name;
        update_dataset_entry($conn, $meta);

        $meta["attribute"] = "sedat_regions";
        $meta["value"] = $sedat_regions;
        update_dataset_entry($conn, $meta);
    
        $meta["attribute"] = "sedat_particles";
        $meta["value"] = $sedat_particles;
        update_dataset_entry($conn, $meta);
    
        $meta["attribute"] = "sedat_emin";
        $meta["value"] = $sedat_emin;
        update_dataset_entry($conn, $meta);
    
        $meta["attribute"] = "sedat_emax";
        $meta["value"] = $sedat_emax;
        update_dataset_entry($conn, $meta);
    
        // Insert the platform.
        $global = array();
        $global["dataset"]=$dataset_id;
        $global["attribute"]="platform";
        $global["entry_number"]=1;
        $global["data_type"]="CDF_CHAR";
        $global["value"]=$platform;
        insert_dataset_metadata_entry($conn,$global);
    
        // Insert the platform type.
        $global["attribute"]="platform_type";
        $global["value"]=$platform_type;
        insert_dataset_metadata_entry($conn,$global);
    
        // Insert the instrument.
        $global["attribute"]="instrument";
        $global["value"]=$instrument;
        insert_dataset_metadata_entry($conn,$global);
    }
    
    // Loop over all rows, interpret, and store into database.
    $block="";          // Block flag: head, global, zvariable, ...
    $in_block=FALSE;    // True when in a block.
    $nrv_index=0;       // Index used to count nrv values when they exist.
    foreach($row_all as $row) {
        // Get lines starting with "#" indicating start of block.
        if(preg_match("/#(\w+)/",$row,$match)>0) {
            $block=$match{1};
            #But still not in block.
            $in_block=FALSE;
        }
        // Read data if in a block.
        if($in_block) {
            // Check type of block.
            switch($block) {
                // header block.
                case "header":
                    // Get tokens separated by ":".
                    $meta=array();
                    $meta["dataset"]=$dataset_id;
                    $meta["attribute"]=strtolower(trim(strtok($row,":")));
                    $meta["value"]=trim(strtok(":"));
                    if(strlen($meta["value"])>0 & strcasecmp($meta["attribute"],"cdf name")!=0) {
                        if($test_mode) {
                            fwrite(STDOUT,"================\nHeader block\n-------------\n");
                            print_r($meta);
                        }
                        else {
                            update_dataset_entry($conn,$meta);
                        }
                    }
                    // Get the variables count line.
                    if(preg_match_all('/(\d+)\/(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\/z\s+(\d+)(.*)/',$row,$match)) {
                        $variable_count=array();
                        $variable_count["rVariables"]=$match[1][0];
                        $variable_count["zVariables"]=$match[2][0];
                        $variable_count["gAttributes"]=$match[3][0];
                        $variable_count["vAttributes"]=$match[4][0];
                        $variable_count["records"]=$match[5][0];
                        $variable_count["dims"]=$match[6][0];
                        if(isset($match[7][0])) {
                            $variable_count["sizes"]=trim($match[7][0]);
                        }
                        else {
                            $variable_count["sizes"]=NULL;
                        }
                        if($test_mode) {
                            fwrite(STDOUT,"===================\nVariable count line\n=============\n");
                            print_r($variable_count);
                        }
                    }
                    break;
                // Global attributes block.
                case "GLOBALattributes":
                    // Extract the global attributes.
                    // If the row starts with a digit then it is the same
                    // attribute with a new entry number.
                    if(preg_match_all('/^(\d+):\s+(.+)\s+\{\s+(.*)\s+\}/',$row,$match)) {
                        $global["entry_number"]=$match[1][0];
                        $global["data_type"]=$match[2][0];
                        $global["value"]=$match[3][0];
                    }
                    else {
                        preg_match_all('/"(.+)"\s+(\d+):\s+(.+)\s+\{\s+(.*)\s+\}/',$row,$match);
                        $global["attribute"]=$match[1][0];
                        $global["entry_number"]=$match[2][0];
                        $global["data_type"]=$match[3][0];
                        $global["value"]=$match[4][0];
                    }
                    if($test_mode) {
                        fwrite(STDOUT,"================\nGlobal attributes block\n-------------\n");
                        print_r($global);
                    }
                    else {
                        insert_dataset_metadata_entry($conn,$global);
                    }
                    break;
                // rVariable block.
                case "variables":
                    // rVariable.
                    if(preg_match_all('/"(.+)"\s+(\w+)\s+(\d+)\s+(\w)(.*)/',$row,$match)) {
                        $var = array();
                        $var["dataset"]=$dataset_id;
                        $var["name"]=$match[1][0];
                        $var["data_type"]=$match[2][0];
                        $var["number_elements"]=$match[3][0];
                        $var["record_variance"]=$match[4][0];
                        $dimvar = trim($match[5][0]);
                        // If there are one or more 'T' use the parsed dims and sizes.
                        if(preg_match("/T/",$dimvar)==1) {
                            $var["dimension_variances"]=str_replace(" ",",",$dimvar);
                            $var["dims"]=$variable_count["dims"];
                            $var["sizes"]=$variable_count["sizes"];
                        }
                        // Else there are only 'F' set dims=0.
                        else {
                            $var["dimension_variances"]=$dimvar;
                            $var["dims"] = 0;
                            $var["sizes"] = "";
                        }
                        if($test_mode) {
                            fwrite(STDOUT,"================\nrVariable block\n-------------\n");
                            print_r($var);
                        }
                        else {
                            insert_variable_metadata_entry($conn,$var);
                        }
                        $nrv_index=0;
                    }
                    // Get variable attributes.
                    if(preg_match_all('/("\w+")\s+(\w+)\s+\{\s+(.*)\s+\}/',$row,$match)) {
                        $varatt = array();
                        $varatt["variable_metadata"]=$var["name"];
                        $varatt["attribute"]=$match[1][0];
                        $varatt["data_type"]=$match[2][0];
                        $varatt["value"]=$match[3][0];
                        if($test_mode) {
                            fwrite(STDOUT,"================\nVariable attributes block\n-------------\n");
                            print_r($varatt);
                        }
                        else {
                            insert_variable_attribute_entry($conn,$varatt,$dataset_id);
                        }
                    }
                    // Get nrv values.
                    if(preg_match_all('/^\[(.*)\]\s*=\s*\{*\s*(.*)\s*\}*/',$row,$match)) {
                        // Remove a possible trailing "}".
                        $nrv_index++;
                        $nrv_pos=trim($match[1][0]," }");
                        $nrv_value=trim($match[2][0]," }");
                        if($test_mode) {
                            fwrite(STDOUT,"================\nNRV values\n-------------\n");
                            fwrite(STDOUT,$var["name"]." <".$nrv_index."> <".$nrv_pos."> <".$nrv_value.">\n");
                        }
                        else {
                            insert_variable_nrv_entry($conn,$var["name"],$nrv_index,$nrv_pos,$nrv_value,$dataset_id);
                        }
                    }
                    break;
                // zVariable block.
                case "zVariables":
                    // Variable name row always contain "Number Elements" and "Dims".
                    if(preg_match_all('/"(.+)"\s+(\w+)\s+(\d+)\s+(\d+)\s+(.*)/',$row,$match)) {
                        $var = array();
                        $var["dataset"]=$dataset_id;
                        $var["name"]=$match[1][0];
                        $var["data_type"]=$match[2][0];
                        $var["number_elements"]=$match[3][0];
                        $var["dims"]=$match[4][0];
                        $rest=$match[5][0];
                        if($var["dims"]>0) {
                            $var["sizes"]="";
                            for($i=0;$i<$var["dims"];$i++) {
                                preg_match_all('/(\d+)\s+(.*)/',$rest,$match);
                                $var["sizes"] .= $match[1][0].",";
                                $rest=$match[2][0];
                            }
                            $var["sizes"]=trim($var["sizes"],",");
                            $rest=$match[2][0];
                        }
                        else {
                            $var["sizes"]="";
                        }
                        if(preg_match_all('/(\w)\s+(.*)/',$rest,$match)) {
                            $var["record_variance"]=$match[1][0];
                            $var["dimension_variances"]=str_replace(" ",",",trim($match[2][0]));
                        }
                        else {
                            $var["record_variance"]=$rest;
                            $var["dimension_variances"]="";
                        }
                        if($test_mode) {
                            fwrite(STDOUT,"================\nzVariable block\n-------------\n");
                            print_r($var);
                        }
                        else {
                            insert_variable_metadata_entry($conn,$var);
                        }
                        $nrv_index=0;
                    }
                    // Get variable attributes.
                    if(preg_match_all('/("\w+")\s+(\w+)\s+\{\s+(.*)\s+\}/',$row,$match)) {
                        $varatt = array();
                        $varatt["variable_metadata"]=$var["name"];
                        $varatt["attribute"]=$match[1][0];
                        $varatt["data_type"]=$match[2][0];
                        $varatt["value"]=$match[3][0];
                        if($test_mode) {
                            fwrite(STDOUT,"================\nVariable attributes block\n-------------\n");
                            print_r($varatt);
                        }
                        else {
                            insert_variable_attribute_entry($conn,$varatt,$dataset_id);
                        }
                    }
                    // Get nrv values.
                    if(preg_match_all('/^\[(.*)\]\s*=\s*\{*\s*(.*)\s*\}*/',$row,$match)) {
                        // Remove a possible trailing "}".
                        $nrv_index++;
                        $nrv_pos=trim($match[1][0]," }");
                        $nrv_value=trim($match[2][0]," }");
                        if($test_mode) {
                            fwrite(STDOUT,"================\nNRV values\n-------------\n");
                            fwrite(STDOUT,$var["name"]." <".$nrv_index."> <".$nrv_pos."> <".$nrv_value.">\n");
                        }
                        else {
                            insert_variable_nrv_entry($conn,$var["name"],$nrv_index,$nrv_pos,$nrv_value,$dataset_id);
                        }
                    }
                    break;
            } // End switch.
        }
        $in_block=TRUE;
    }

    if(!$test_mode) {
        // Create dataset table.
        create_dataset_table($conn,$dataset_id,$dataset_name);
    }

    fwrite(STDOUT,"Done!\n");

}


/**
* Produce a list of files already ingested for the dataset.
*
* @param mysqli $link    A MySQL database connection
* @param int    $tableid The dataset id
* @return array A multidim array with id, file names, and file dates
*/
function tfileList($link, $tableid)
{
    $files = array();
    $query = "SELECT * FROM dataset_file WHERE dataset_id=" . $tableid;
    $result = mysqli_query($link, $query);

    while ($row = mysqli_fetch_array($result, MYSQLI_ASSOC))
    {
        $files["id"][] = $row["id"];
        $files["dfname"][] = $row["filename"];
        $files["fdate"][] = $row["filedate"];
    }
    mysqli_free_result($result);

    return $files;
}

/**
* Update an entry in the dataset table.
*
* The <code>$meta</code> argument is an array with the following structure:
* <code>
*   $meta = array (
*     "attribute" => <column name in dataset>,
*     "value"     => <the new value>,
*     "dataset"   => <the id of the row to be updated>
    )
* </code>
*
* @param mysqli $conn  A MySQL database connection.
* @param array  $meta  An array with the new entry values.
*/
function update_dataset_entry($conn,$meta) {
    $q = 'UPDATE dataset SET '.no_quote(s_to_us($meta["attribute"])).'='.quote($meta["value"]);
    $q .= ' WHERE dataset.id='.$meta["dataset"];
    if (!($conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
}

?>