;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Read in an ODI dataset
;
; Author: D. Heynderickx (DH Consultancy)
; Created: 25 Nov 2009
;
; INPUTS
;   datasetname: STRING; datasetname without 'dataset_', e.g. 'azur_ei_88'
;
; OUTPUTS
;   data: structure containing all data (RV and NRV), plus Julian day
;         numbers if argument jd is set;
;   query: string scalar containing the SQL query used to retrieve the data
;
; Optional arguments
;   global_att: structure containing the dataset global attributes
;   var_att: structure containing the variable attributes
;   var_info: structure containing the IDL CDF variable info
;   progress: not used, but kept for backward compatibility
;   jd: flag to add Julian day numbers to the data structure
;   vary_rec: structure array containing only the RV variables
;   novary_rec: structure scalar containing the NRV variables
;   from: string scalar with start epoch in SQL DATETIME format
;   to: string scalar with end epoch in SQL DATETIME format
;   sql_where: string scalar containing optional SQL WHERE clauses
;              should not contain 'WHERE' or leading 'AND' or 'OR' etc.
;
; Requires odi_tools to be compiled
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

PRO read_odi, datasetname, data, query, global_att=global_att, $
              var_att=var_att, var_info=var_info, progress=progress, $
              jd=jd, vary_rec=vary_rec, novary_rec=novary_rec, $
              from=from, to=to, sql_where=sql_where

; Connect to the ODI database
  DBConnect, oDB

; Retrieve the dataset id and COL or ROW majority
  query = "SELECT id, LEFT(majority, 20) AS majority FROM dataset " + $
          "WHERE name='dataset_" + datasetname + "'"
  oRS = OBJ_NEW('IDLDBRecordSet', oDB, SQL=query)
  IF (oRS->CurrentRecord() LT 0) THEN BEGIN
    PRINT, 'Failed to open record set for dataset id'
    RETURN
  ENDIF
  dataset = oRS->GetRecord()
  OBJ_DESTROY, oRS

; This defines the variables to be read
; '*' means all variables
; Variable selection is made by replacing with a string array of variable names
  varnames = '*'

; Set up a structure containing the variable definitions
  vars = odi_vardefs(varnames, dataset.id, oDB)

  numcols = N_ELEMENTS(vars)

  var_info = 0
  data = 0
  FOR i=0,numcols-1 DO BEGIN
; Add variable declaration and SQL strings to structure vars
    odi_descvar, vars[i], dataset.majority, vtype, sql
    vars[i].vtype = vtype
    vars[i].sql = sql
; Construct the data structure
    res = EXECUTE("dvar = " + vars[i].vtype)
    dvar = cast_odi_idl(dvar, vars[i].idl_type)
    IF (SIZE(data, /TYPE) NE 8) THEN $
      data = CREATE_STRUCT(vars[i].name, dvar) $
    ELSE $
      data = CREATE_STRUCT(data, vars[i].name, dvar)
; Construct the var_info structure
    varinq = { IS_ZVAR: 1, NAME: vars[i].name, $
               DATATYPE: vars[i].cdf_type, $
               NUMELEM: LONG(vars[i].number_elements), $
               RECVAR: (vars[i].rvar EQ 'T') ? "VARY" : "NOVARY", $
               DIMVAR: BYTE(STRSPLIT(vars[i].dvar, ",", /EXTRACT) EQ 'T'), $
               DIM: LONG(vars[i].dimensions) }
    IF (SIZE(var_info, /TYPE) NE 8) THEN $
      var_info = CREATE_STRUCT(vars[i].name, varinq) $
    ELSE $
      var_info = CREATE_STRUCT(var_info, vars[i].name, varinq)
  ENDFOR
; Add the Julian day field if requested.
  IF (KEYWORD_SET(jd) AND ~IS_IN('JULDAY', TAG_NAMES(data))) THEN $
    data = CREATE_STRUCT(data, 'JULDAY', [0.0D0])

; Retrieve and store global attributes from table dataset_metadata
  IF ARG_PRESENT(global_att) THEN BEGIN
    sql = "SELECT LEFT(dm.attribute, 100) AS name, " + $
          "LEFT(dt.idl_type, 20) AS type, dm.entry_number, " + $
          "LEFT(dm.value, 1000) AS value FROM dataset_metadata AS dm " + $
          "LEFT JOIN data_type AS dt ON (dm.data_type_id=dt.id) " + $
          "WHERE dm.dataset_id=" + STRING(dataset.id) + " AND " + $
          "dm.attribute!='platform' AND dm.attribute!='platform_type' " + $
          "AND dm.attribute!='instrument'"
    oRS = OBJ_NEW('IDLDBRecordSet', oDB, SQL=sql)
    IF (oRS->CurrentRecord() GE 0) THEN BEGIN
      att = oRS->GetRecord()
      WHILE (oRS->MoveCursor(/NEXT)) DO $
        att = [att, oRS->GetRecord()]
      OBJ_DESTROY, oRS
    ENDIF ELSE BEGIN
      PRINT, 'Error reading global attributes!'
      RETURN
    ENDELSE

    global_att = 0
    FOR i=0,N_ELEMENTS(att)-1 DO BEGIN
      IF (SIZE(global_att, /TYPE) NE 8) THEN BEGIN
        global_att = CREATE_STRUCT(att[i].name, $
                                   cast_odi_idl(att[i].value, att[i].type))
      ENDIF ELSE BEGIN
        IF (att[i].entry_number GT 1) THEN $
          global_att.(N_TAGS(global_att)-1) += " " + cast_odi_idl(att[i].value, att[i].type) $
        ELSE $
          global_att = CREATE_STRUCT(global_att, att[i].name, $
                                     cast_odi_idl(att[i].value, att[i].type))
      ENDELSE
    ENDFOR
  ENDIF

; Retrieve variable attributes from table variable_attribute
  var_att = 0
  IF ARG_PRESENT(var_att) THEN BEGIN
    FOR i=0,numcols-1 DO BEGIN
      sql = "SELECT LEFT(va.name, 50) AS name, " + $
            "LEFT(va.value, 5000) AS value, " + $
            "LEFT(dt.idl_type, 20) AS idl_type FROM " + $
            "variable_attribute AS va LEFT JOIN data_type AS dt " + $
            "ON (va.data_type_id=dt.id) " + $
            "WHERE va.variable_metadata_id=" + STRING(vars[i].id)
      oRS = OBJ_NEW('IDLDBRecordSet', oDB, SQL=sql)
      IF (oRS->CurrentRecord() GE 0) THEN BEGIN
        att = oRS->GetRecord()
        WHILE (oRS->MoveCursor(/NEXT)) DO $
          att = [att, oRS->GetRecord()]
        OBJ_DESTROY, oRS
      ENDIF ELSE BEGIN
        PRINT, ' Error reading variable attributes for ' + vars[i].name + '!'
        RETURN
      ENDELSE
      vatts = 0
      FOR j=0,N_ELEMENTS(att)-1 DO BEGIN
        IF (SIZE(vatts, /TYPE) NE 8) THEN $
          vatts = CREATE_STRUCT(att[j].name, $
                                cast_odi_idl(att[j].value, att[j].idl_type)) $
        ELSE $
          vatts = CREATE_STRUCT(vatts, att[j].name, $
                                cast_odi_idl(att[j].value, att[j].idl_type))
      ENDFOR
      IF (SIZE(var_att, /TYPE) NE 8) THEN $
        var_att = CREATE_STRUCT(vars[i].name, vatts) $
      ELSE $
        var_att = CREATE_STRUCT(var_att, vars[i].name, vatts)
    ENDFOR
  ENDIF

; Create the SQL query to retrieve the data
  query = odi_rvsql(vars) + " FROM dataset_" + datasetname
  wsql = ""
  IF (N_ELEMENTS(from) GT 0) THEN $
    wsql = [wsql,"cdf_epoch>=" + STRING(sql_date2cdf_epoch(from), FORMAT='(F20.1)')]
  IF (N_ELEMENTS(to) GT 0) THEN $
    wsql = [wsql,"cdf_epoch<=" + STRING(sql_date2cdf_epoch(to), FORMAT='(F20.1)')]
  IF (N_ELEMENTS(sql_where) GT 0) THEN wsql = [wsql,sql_where]

  wquery = ""
  IF (N_ELEMENTS(wsql) GT 1) THEN BEGIN
    wquery += " WHERE"
    FOR i=1,N_ELEMENTS(wsql)-1 DO BEGIN
      IF (i EQ 1) THEN $
        wquery += " " + wsql[i] $
      ELSE $
        wquery += " AND " + wsql[i]
    ENDFOR
  ENDIF
  query += wquery + " ORDER BY cdf_epoch ASC"

; Obtain the number or records
  sql = "SELECT CAST(count(cdf_epoch) AS BINARY) AS numrecs FROM " + $
        "dataset_" + datasetname + wquery
  oRS = OBJ_NEW('IDLDBRecordSet', oDB, SQL=sql)
  IF (OBJ_VALID(oRS)) THEN BEGIN
    rec = oRS->GetRecord()
    numents = LONG(rec.numrecs)
  ENDIF ELSE BEGIN
    numents = 0L
  ENDELSE
  OBJ_DESTROY, oRS

; Read in the RV variables
  IF (numents GT 0) THEN BEGIN
    oRS = OBJ_NEW('IDLDBRecordSet', oDB, N_BUFFERS=10000L, SQL=query)
    status = (oRS->CurrentRecord() GE 0)
    IF (status LT 0) THEN BEGIN
      PRINT, 'Failed to open record set in read_odi'
      RETURN
    ENDIF
    rec = oRS->GetRecord()
    rec = REPLICATE(rec, numents)
    FOR i=2L,numents DO BEGIN
      status = oRS->MoveCursor(/NEXT)
      IF (status LE 0) THEN BEGIN
        PRINT, 'Failed to advance record set pointer in read_odi'
        RETURN
      ENDIF
      rec[i-1] = oRS->GetRecord()
    ENDFOR
    OBJ_DESTROY, oRS
    data = REPLICATE(data, numents)
  ENDIF ELSE BEGIN
    numents = 0
    PRINT, 'No records in the dataset match the query'
  ENDELSE

; Load the data from the database retrieved records and NRV variables
  nr = 0L
  FOR itag=0,numcols-1 DO BEGIN
    IF (vars[itag].rvar EQ 'T') THEN BEGIN
; Copy RV values
      FOR idim=0,vars[itag].dimensions[0]-1 DO BEGIN
        IF (numents GT 0) THEN data.(itag)[idim] = rec.(nr)
        nr++  ; move on to the next field in the SQL dataset record
      ENDFOR
    ENDIF ELSE BEGIN
; Read NRV values
      sql = "SELECT LEFT(value, 5000) AS value FROM variable_nrv " + $
            "WHERE variable_metadata_id=" + STRING(vars[itag].id) + $
            " ORDER BY ind ASC"
      oRS = OBJ_NEW('IDLDBRecordSet', oDB, SQL=sql)
      IF (oRS->CurrentRecord() GE 0) THEN BEGIN
        nrv = oRS->GetRecord()
        WHILE (oRS->MoveCursor(/NEXT)) DO $
          nrv = [nrv, oRS->GetRecord()]
        OBJ_DESTROY, oRS
      ENDIF ELSE BEGIN
        PRINT, ' Error reading NRV variables!'
        RETURN
      ENDELSE
      FOR idim=0,vars[itag].dimensions[0]-1 DO $
        data.(itag)[idim] = nrv[idim].value
    ENDELSE
  ENDFOR

; If Julian day calculation required then calculate from Epoch
  IF (KEYWORD_SET(jd) AND (numents GT 0)) THEN BEGIN
; Use a reasonably current time, then Julian day is the
; #ms/()#ms/day) + JD of that base time.
    cdf_epoch, base_epoch, 2000, 1, 1, 0, 0, 0, 0.0, /COMPUTE
    data.julday = (data.epoch-base_epoch) / 8.64D7 + JULDAY(1, 1, 2000, 0, 0, 0.0D0)
  ENDIF

; Now build the (no_)vary_rec structures from the data structure.
  IF ARG_PRESENT(vary_rec) THEN $
    vary_rec = read_cdf_make_rec_vary_str(data, var_info)
  IF ARG_PRESENT(novary_rec) THEN $
    novary_rec = read_cdf_make_rec_vary_str(data, var_info, /NOVARY)

  OBJ_DESTROY, oDB

END